#!/usr/bin/env python
# -*- coding: utf-8 -*-

# dialogs.roi_map.py
"""
GUI tools to edit the ROI Name map
"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

import wx
from functools import partial
from dvha.models.data_table import DataTable
from dvha.tools.errors import ROIVariationErrorDialog, ROIVariationError
from dvha.tools.utilities import get_selected_listctrl_items, MessageDialog
from dvha.tools.roi_map_generator import ROIMapGenerator
from dvha.tools.roi_name_manager import clean_name

try:
    from dvha.tools.name_prediction import ROINamePredictor
except ImportError:
    ROINamePredictor = None


class AddPhysician(wx.Dialog):
    """
    Create a new physician in the ROI map with ability to copy the initial map from a current physician
    """

    def __init__(self, roi_map, initial_physician=None):
        """
        :param roi_map: the roi_map object
        :type roi_map: DatabaseROIs
        :param initial_physician: optionally set the new physician name in the text_ctrl
        :type initial_physician: str
        """
        wx.Dialog.__init__(self, None)

        self.roi_map = roi_map
        physicians = ["NONE"] + [
            p for p in list(self.roi_map.physicians) if p != "DEFAULT"
        ]
        if initial_physician and initial_physician.lower() not in {"default"}:
            self.initial_physician = initial_physician
        else:
            self.initial_physician = ""

        self.text_ctrl_physician = wx.TextCtrl(self, wx.ID_ANY, "")
        self.combo_box_copy_from = wx.ComboBox(
            self,
            wx.ID_ANY,
            choices=physicians,
            style=wx.CB_DROPDOWN | wx.CB_READONLY,
        )
        self.combo_box_copy_from.SetValue(
            physicians[0]
        )  # MSW needs this set explicitly
        self.checkbox_variations = wx.CheckBox(
            self, wx.ID_ANY, "Include Variations"
        )
        self.button_ok = wx.Button(self, wx.ID_OK, "OK")
        self.button_cancel = wx.Button(self, wx.ID_CANCEL, "Cancel")

        self.__set_properties()
        self.__do_layout()
        self.__do_bind()

        self.res = None

        self.run()

    def __do_bind(self):
        self.Bind(
            wx.EVT_TEXT,
            self.update_enable,
            id=self.text_ctrl_physician.GetId(),
        )
        self.Bind(
            wx.EVT_COMBOBOX,
            self.update_enable,
            id=self.combo_box_copy_from.GetId(),
        )

    def __set_properties(self):
        self.SetTitle("Add Physician to ROI Map")
        self.checkbox_variations.SetValue(1)
        self.text_ctrl_physician.SetValue(self.initial_physician)
        self.update_enable()

    def __do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_ok_cancel = wx.BoxSizer(wx.HORIZONTAL)
        sizer_widgets = wx.StaticBoxSizer(
            wx.StaticBox(self, wx.ID_ANY, ""), wx.HORIZONTAL
        )
        sizer_copy_from = wx.BoxSizer(wx.VERTICAL)
        sizer_new_physician = wx.BoxSizer(wx.VERTICAL)

        label_new_physician = wx.StaticText(self, wx.ID_ANY, "New Physician:")
        sizer_new_physician.Add(
            label_new_physician, 0, wx.EXPAND | wx.LEFT | wx.RIGHT | wx.TOP, 5
        )
        sizer_new_physician.Add(
            self.text_ctrl_physician,
            0,
            wx.BOTTOM | wx.EXPAND | wx.LEFT | wx.RIGHT,
            5,
        )
        sizer_widgets.Add(sizer_new_physician, 1, wx.EXPAND | wx.TOP, 10)

        label_copy_from = wx.StaticText(self, wx.ID_ANY, "Copy From:")
        sizer_copy_from.Add(
            label_copy_from, 0, wx.EXPAND | wx.LEFT | wx.RIGHT | wx.TOP, 5
        )
        sizer_copy_from.Add(
            self.combo_box_copy_from, 0, wx.EXPAND | wx.LEFT | wx.RIGHT, 5
        )
        sizer_widgets.Add(sizer_copy_from, 1, wx.EXPAND | wx.TOP, 10)
        sizer_widgets.Add(self.checkbox_variations, 0, wx.EXPAND | wx.ALL, 10)
        sizer_wrapper.Add(sizer_widgets, 0, wx.ALL | wx.EXPAND, 10)

        sizer_ok_cancel.Add(
            self.button_ok, 0, wx.EXPAND | wx.LEFT | wx.RIGHT, 5
        )
        sizer_ok_cancel.Add(self.button_cancel, 0, wx.LEFT | wx.RIGHT, 5)
        sizer_wrapper.Add(
            sizer_ok_cancel,
            0,
            wx.ALIGN_RIGHT | wx.BOTTOM | wx.LEFT | wx.RIGHT,
            10,
        )

        self.SetSizer(sizer_wrapper)
        self.Layout()
        self.Fit()
        self.Center()

    def action(self):
        if self.combo_box_copy_from.GetValue().lower() == "none":
            # self.roi_map.add_physician(self.text_ctrl_physician.GetValue(), add_institutional_rois=False)
            physician = self.text_ctrl_physician.GetValue()
            dlg = TG263Dialog(physician, self.roi_map)
            if dlg.map_file_path is not None:
                self.roi_map.import_physician_roi_map(dlg.map_file_path)

        else:
            self.roi_map.copy_physician(
                self.text_ctrl_physician.GetValue(),
                copy_from=self.combo_box_copy_from.GetValue(),
                include_variations=self.checkbox_variations.GetValue(),
            )

    def run(self):
        self.res = self.ShowModal()
        if self.res == wx.ID_OK:
            self.action()
        self.Destroy()

    def update_enable(self, *evt):
        invalid_choices = list(self.roi_map.physicians) + [""]
        new = clean_name(self.text_ctrl_physician.GetValue(), physician=True)
        self.button_ok.Enable(new not in invalid_choices)
        disable = (
            self.text_ctrl_physician.GetValue() == ""
            or self.combo_box_copy_from.GetValue().lower() == "none"
        )
        self.checkbox_variations.Enable(not disable)


class TG263Dialog(wx.Dialog):
    def __init__(self, physician, roi_map):
        wx.Dialog.__init__(self, None)

        self.physician = physician
        self.roi_map = roi_map
        self.generator = ROIMapGenerator()

        self.map_file_path = None

        self.checkbox_keys = self.generator.anatomic_groups
        self.checkbox = {
            key: wx.CheckBox(self, wx.ID_ANY, key)
            for key in self.checkbox_keys
        }

        self.button_keys = ["Select All", "Deselect All", "OK", "Cancel"]
        button_ids = [wx.ID_ANY, wx.ID_ANY, wx.ID_OK, wx.ID_CANCEL]
        self.button = {
            key: wx.Button(self, button_ids[i], key)
            for i, key in enumerate(self.button_keys)
        }

        self.__set_properties()
        self.__do_bind()
        self.__do_layout()

        self.run()

    def __set_properties(self):
        self.SetTitle("Create ROI Map")
        self.SetMinSize((300, 100))

    def __do_bind(self):
        self.Bind(
            wx.EVT_BUTTON,
            self.on_select_all,
            id=self.button["Select All"].GetId(),
        )
        self.Bind(
            wx.EVT_BUTTON,
            self.on_deselect_all,
            id=self.button["Deselect All"].GetId(),
        )

    def __do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_main = wx.BoxSizer(wx.HORIZONTAL)
        sizer_body_site = wx.StaticBoxSizer(
            wx.StaticBox(self, wx.ID_ANY, "Body Sites"), wx.VERTICAL
        )
        sizer_buttons = wx.BoxSizer(wx.VERTICAL)

        for key in self.checkbox_keys:
            sizer_body_site.Add(self.checkbox[key], 1, wx.EXPAND, 0)

        sizer_buttons.Add((20, 20), 0, 0, 0)
        for key in self.button_keys:
            sizer_buttons.Add(self.button[key], 1, wx.EXPAND | wx.ALL, 5)

        sizer_main.Add(sizer_body_site, 0, wx.ALL, 5)
        sizer_main.Add(sizer_buttons, 0, wx.LEFT, 10)

        title = wx.StaticText(
            self,
            wx.ID_ANY,
            "Create ROI Map for %s with selected body sites" % self.physician,
        )
        sizer_wrapper.Add(title, 0, wx.EXPAND | wx.ALL, 10)
        sizer_wrapper.Add(sizer_main, 1, wx.ALL, 10)

        self.SetSizer(sizer_wrapper)
        self.Layout()
        self.Fit()
        self.Center()

    def set_checkbox_values(self, value):
        for checkbox in self.checkbox.values():
            checkbox.SetValue(value)

    def on_select_all(self, *evt):
        self.set_checkbox_values(True)

    def on_deselect_all(self, *evt):
        self.set_checkbox_values(False)

    def run(self):
        res = self.ShowModal()
        if res == wx.ID_OK:
            map_file_name = "physician_%s.roi" % clean_name(
                self.physician, physician=True
            )
            self.map_file_path = self.generator(
                map_file_name, body_sites=self.checked_values
            )
        self.Destroy()

    @property
    def checked_values(self):
        return [
            key
            for key, checkbox in self.checkbox.items()
            if checkbox.GetValue()
        ]


class RoiManager(wx.Dialog):
    """
    Dialog accessible from the DICOM import GUI to allow user to easily update the ROI map with new plans
    """

    def __init__(self, parent, roi_map, physician, physician_roi):
        """
        :param parent: GUI parent
        :param roi_map: roi_map object
        :type roi_map: DatabaseROIs
        :param physician: initial physician value
        :type physician: str
        :param physician_roi: initial physician roi
        :type physician_roi: str
        """
        wx.Dialog.__init__(self, parent)
        self.parent = parent
        self.roi_map = roi_map
        self.initial_physician = physician
        self.initial_physician_roi = physician_roi

        self.combo_box_physician = wx.ComboBox(
            self,
            wx.ID_ANY,
            choices=list(self.roi_map.physicians),
            style=wx.CB_DROPDOWN | wx.CB_READONLY,
        )
        self.combo_box_physician_roi = wx.ComboBox(
            self, wx.ID_ANY, choices=[], style=wx.CB_DROPDOWN | wx.CB_READONLY
        )
        self.list_ctrl_variations = wx.ListCtrl(
            self,
            wx.ID_ANY,
            style=wx.LC_NO_HEADER | wx.LC_REPORT | wx.BORDER_SUNKEN,
        )
        self.button_select_all = wx.Button(self, wx.ID_ANY, "Select All")
        self.button_deselect_all = wx.Button(self, wx.ID_ANY, "Deselect All")
        self.button_add = wx.Button(self, wx.ID_ANY, "Add")
        self.button_delete = wx.Button(self, wx.ID_ANY, "Delete")
        self.button_move = wx.Button(self, wx.ID_ANY, "Move")
        self.button_dismiss = wx.Button(self, wx.ID_CANCEL, "Dismiss")

        self.button_move.Disable()
        self.button_delete.Disable()
        self.button_deselect_all.Disable()

        self.button_add_physician = wx.Button(self, wx.ID_ANY, "Add")
        self.button_add_physician_roi = wx.Button(self, wx.ID_ANY, "Add")

        self.columns = ["Variations"]
        self.data_table = DataTable(
            self.list_ctrl_variations, columns=self.columns, widths=[400]
        )

        self.__set_properties()
        self.__do_layout()
        self.__do_bind()

        self.physician_ticker()

        self.run()

    def __set_properties(self):
        self.SetTitle("ROI Manager")
        self.combo_box_physician.SetValue(self.initial_physician)
        self.update_physician_rois()
        self.combo_box_physician_roi.SetValue(self.initial_physician_roi)
        self.update_variations()

    def __do_bind(self):
        self.Bind(
            wx.EVT_COMBOBOX,
            self.physician_ticker,
            id=self.combo_box_physician.GetId(),
        )
        self.Bind(
            wx.EVT_COMBOBOX,
            self.physician_roi_ticker,
            id=self.combo_box_physician_roi.GetId(),
        )
        self.Bind(
            wx.EVT_BUTTON,
            self.add_physician,
            id=self.button_add_physician.GetId(),
        )
        self.Bind(
            wx.EVT_BUTTON,
            self.add_physician_roi,
            id=self.button_add_physician_roi.GetId(),
        )
        self.Bind(
            wx.EVT_BUTTON, self.select_all, id=self.button_select_all.GetId()
        )
        self.Bind(
            wx.EVT_BUTTON,
            self.deselect_all,
            id=self.button_deselect_all.GetId(),
        )
        self.Bind(
            wx.EVT_BUTTON, self.add_variation, id=self.button_add.GetId()
        )
        self.Bind(
            wx.EVT_BUTTON, self.move_variations, id=self.button_move.GetId()
        )
        self.Bind(
            wx.EVT_BUTTON,
            self.delete_variations,
            id=self.button_delete.GetId(),
        )
        self.Bind(
            wx.EVT_LIST_ITEM_SELECTED,
            self.update_button_enable,
            id=self.list_ctrl_variations.GetId(),
        )
        self.Bind(
            wx.EVT_LIST_ITEM_DESELECTED,
            self.update_button_enable,
            id=self.list_ctrl_variations.GetId(),
        )

    def __do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_buttons = wx.BoxSizer(wx.HORIZONTAL)
        sizer_variation_buttons = wx.BoxSizer(wx.VERTICAL)
        sizer_variation_table = wx.BoxSizer(wx.VERTICAL)
        sizer_select = wx.StaticBoxSizer(
            wx.StaticBox(self, wx.ID_ANY, ""), wx.VERTICAL
        )
        sizer_variations = wx.BoxSizer(wx.HORIZONTAL)
        sizer_physician_roi = wx.BoxSizer(wx.VERTICAL)
        sizer_physician_roi_row_2 = wx.BoxSizer(wx.HORIZONTAL)
        sizer_physician = wx.BoxSizer(wx.VERTICAL)

        label_physician = wx.StaticText(self, wx.ID_ANY, "Physician:")
        sizer_physician.Add(label_physician, 0, wx.LEFT, 5)
        sizer_physician_row = wx.BoxSizer(wx.HORIZONTAL)
        sizer_physician_row.Add(
            self.combo_box_physician, 1, wx.EXPAND | wx.LEFT | wx.RIGHT, 5
        )
        sizer_physician_row.Add(
            self.button_add_physician, 0, wx.EXPAND | wx.LEFT | wx.RIGHT, 10
        )
        sizer_physician.Add(sizer_physician_row, 1, wx.EXPAND, 0)
        sizer_select.Add(sizer_physician, 0, wx.ALL | wx.EXPAND, 5)

        label_physician_roi = wx.StaticText(self, wx.ID_ANY, "Physician ROI:")
        sizer_physician_roi.Add(label_physician_roi, 0, wx.LEFT, 5)
        sizer_physician_roi_row_2.Add(
            self.combo_box_physician_roi, 1, wx.EXPAND | wx.LEFT | wx.RIGHT, 5
        )
        sizer_physician_roi_row_2.Add(
            self.button_add_physician_roi,
            0,
            wx.EXPAND | wx.LEFT | wx.RIGHT,
            10,
        )
        sizer_physician_roi.Add(sizer_physician_roi_row_2, 0, wx.EXPAND, 0)
        sizer_select.Add(sizer_physician_roi, 0, wx.ALL | wx.EXPAND, 5)

        label_variations = wx.StaticText(self, wx.ID_ANY, "Variations:")
        label_variations_buttons = wx.StaticText(self, wx.ID_ANY, " ")
        sizer_variation_table.Add(label_variations, 0, 0, 0)
        sizer_variation_table.Add(
            self.list_ctrl_variations, 1, wx.ALL | wx.EXPAND, 0
        )
        sizer_variations.Add(sizer_variation_table, 0, wx.ALL, 5)
        sizer_variation_buttons.Add(label_variations_buttons, 0, 0, 0)
        sizer_variation_buttons.Add(
            self.button_add, 0, wx.EXPAND | wx.LEFT | wx.RIGHT | wx.BOTTOM, 5
        )
        sizer_variation_buttons.Add(
            self.button_delete, 0, wx.EXPAND | wx.ALL, 5
        )
        sizer_variation_buttons.Add(self.button_move, 0, wx.EXPAND | wx.ALL, 5)
        sizer_variation_buttons.Add(
            self.button_select_all, 0, wx.EXPAND | wx.ALL, 5
        )
        sizer_variation_buttons.Add(
            self.button_deselect_all, 0, wx.EXPAND | wx.ALL, 5
        )
        sizer_variations.Add(sizer_variation_buttons, 0, wx.EXPAND | wx.ALL, 5)
        sizer_select.Add(sizer_variations, 0, wx.ALL | wx.EXPAND, 5)

        sizer_wrapper.Add(sizer_select, 0, wx.ALL, 5)
        sizer_buttons.Add(self.button_dismiss, 0, wx.ALIGN_CENTER | wx.ALL, 5)
        sizer_wrapper.Add(sizer_buttons, 0, wx.ALIGN_CENTER | wx.ALL, 0)

        self.SetSizer(sizer_wrapper)
        self.Layout()
        self.Fit()
        self.Center()

    def run(self):
        self.ShowModal()
        self.Destroy()

    def update_physicians(self, old_physicians=None):
        choices = list(self.roi_map.physicians)
        self.update_combo_box_choices(
            self.combo_box_physician, choices, old_physicians
        )

    def update_physician_rois(self, old_physician_rois=None):
        choices = self.roi_map.get_physician_rois(self.physician)
        self.update_combo_box_choices(
            self.combo_box_physician_roi, choices, old_physician_rois
        )

    @staticmethod
    def update_combo_box_choices(combo_box, choices, old_choices):
        if choices:
            value = choices[0]
            if old_choices:
                value = list(set(choices) - set(old_choices))
                value = value[0] if len(value) else combo_box.GetValue()

            combo_box.Clear()
            combo_box.AppendItems(choices)
            combo_box.SetValue(value)

    def update_variations(self):
        self.data_table.set_data(self.variation_table_data, self.columns)
        self.update_button_enable()

    def physician_ticker(self, *evt):
        self.update_physician_rois()
        self.update_variations()
        if self.physician == "DEFAULT":
            self.set_all_button_enable(False)
        else:
            self.set_all_button_enable(True)
            self.update_button_enable()

    def physician_roi_ticker(self, evt):
        self.update_variations()

    @property
    def physician(self):
        return self.combo_box_physician.GetValue()

    @property
    def physician_roi(self):
        return self.combo_box_physician_roi.GetValue()

    @property
    def variations(self):
        variations = self.roi_map.get_variations(
            self.combo_box_physician.GetValue(),
            self.combo_box_physician_roi.GetValue(),
        )
        variations = list(
            set(variations) - {self.combo_box_physician_roi.GetValue()}
        )  # remove physician roi
        variations.sort()
        return variations

    @property
    def variation_table_data(self):
        return {"Variations": self.variations}

    @property
    def selected_indices(self):
        return get_selected_listctrl_items(self.list_ctrl_variations)

    @property
    def selected_values(self):
        return [
            self.list_ctrl_variations.GetItem(i, 0).GetText()
            for i in self.selected_indices
        ]

    @property
    def variation_count(self):
        return len(self.variations)

    def select_all(self, evt):
        self.apply_global_selection()

    def deselect_all(self, evt):
        self.apply_global_selection(on=0)

    def apply_global_selection(self, on=1):
        for i in range(self.variation_count):
            self.list_ctrl_variations.Select(i, on=on)

    def delete_variations(self, evt):
        self.roi_map.delete_variations(
            self.physician, self.physician_roi, self.selected_values
        )
        self.update_variations()

    def add_variation(self, evt):
        AddVariation(
            self.parent, self.physician, self.roi_map, self.physician_roi
        )
        self.update_variations()

    def move_variations(self, evt):
        choices = [
            roi
            for roi in self.roi_map.get_physician_rois(self.physician)
            if roi != self.physician_roi
        ]
        MoveVariationDialog(
            self,
            self.selected_values,
            self.physician,
            self.physician_roi,
            choices,
            self.roi_map,
        )
        self.update_variations()

    def update_button_enable(self, *evt):
        if self.selected_indices:
            self.button_move.Enable()
            self.button_delete.Enable()
            self.button_deselect_all.Enable()
        else:
            self.button_move.Disable()
            self.button_delete.Disable()
            self.button_deselect_all.Disable()

    def set_all_button_enable(self, status):
        self.button_select_all.Enable(status)
        self.button_deselect_all.Enable(status)
        self.button_add.Enable(status)
        self.button_delete.Enable(status)
        self.button_move.Enable(status)
        # self.button_add_physician.Enable(status)
        self.button_add_physician_roi.Enable(status)

    def add_physician_roi(self, evt):
        old_physician_rois = self.roi_map.get_physician_rois(self.physician)
        AddPhysicianROI(self.parent, self.physician, self.roi_map)
        self.update_physician_rois(old_physician_rois=old_physician_rois)

    def add_physician(self, evt):
        old_physicians = list(self.roi_map.physicians)
        AddPhysician(self.roi_map)
        self.update_physicians(old_physicians=old_physicians)


class AddVariationDlg(wx.Dialog):
    """
    Generic class for Add or Assign new variation for a specified physician and physician roi
    """

    def __init__(self, parent, physician, roi_map, input_roi_name, mode="add"):
        wx.Dialog.__init__(self, parent)
        self.parent = parent
        self.physician = physician
        self.input_roi_name = input_roi_name
        self.roi_map = roi_map
        self.mode = mode

        self.label_input = wx.StaticText(
            self,
            wx.ID_ANY,
            ["Physician ROI:", "New variation:"][self.mode == "add"],
        )
        if mode == "add":
            self.user_input = wx.TextCtrl(self, wx.ID_ANY, "")
        else:
            choices = roi_map.get_physician_rois(physician)
            self.user_input = wx.ComboBox(
                self,
                wx.ID_ANY,
                choices=choices,
                style=wx.CB_DROPDOWN | wx.CB_READONLY,
            )
            self.prediction_label = wx.StaticText(self, wx.ID_ANY, "")

        self.button_ok = wx.Button(self, wx.ID_OK, "Add")
        self.button_cancel = wx.Button(self, wx.ID_CANCEL, "Cancel")

        if mode == "assign":
            self.predict_physician_roi()

        self.__do_bind()
        self.__set_properties()
        self.__do_layout()

        self.run()

    def predict_physician_roi(self):
        if ROINamePredictor is not None:
            predictor = ROINamePredictor(self.roi_map)
            prediction = predictor.get_best_roi_match(
                self.input_roi_name, self.physician, return_score=True
            )
            if prediction:
                self.user_input.SetValue(prediction[0])
                self.prediction_label.SetLabelText(
                    "Variation: %s\nPrediction: %s\nScore: %0.0f%%"
                    % (self.input_roi_name, prediction[0], prediction[1])
                )

    def __do_bind(self):
        self.Bind(
            [wx.EVT_COMBOBOX, wx.EVT_TEXT][self.mode == "add"],
            self.enable_add_button,
            id=self.user_input.GetId(),
        )

    def __set_properties(self):
        self.SetTitle(
            "%s %s as a variation for %s"
            % (self.mode.capitalize(), self.input_roi_name, self.physician)
        )
        self.user_input.SetMinSize((400, self.user_input.GetSize()[1]))
        self.button_ok.SetToolTip(
            "If Add is disabled, requested variation is already in use by %s."
            % self.physician
        )

    def __do_layout(self):
        sizer_frame = wx.BoxSizer(wx.VERTICAL)
        sizer_buttons = wx.BoxSizer(wx.HORIZONTAL)
        sizer_variation = wx.StaticBoxSizer(
            wx.StaticBox(self, wx.ID_ANY, ""), wx.VERTICAL
        )

        sizer_variation.Add(self.label_input, 0, 0, 0)
        sizer_variation.Add(self.user_input, 0, wx.EXPAND, 0)
        sizer_frame.Add(sizer_variation, 0, wx.EXPAND, 0)

        if self.mode == "assign":
            sizer_variation.Add(self.prediction_label, 0, 0, 0)

        sizer_buttons.Add(self.button_ok, 1, wx.ALL | wx.EXPAND, 5)
        sizer_buttons.Add(self.button_cancel, 1, wx.ALL | wx.EXPAND, 5)
        sizer_frame.Add(sizer_buttons, 0, wx.ALL | wx.EXPAND, 5)

        self.SetSizer(sizer_frame)
        sizer_frame.Fit(self)
        self.Layout()
        self.Center()

    def enable_add_button(self, evt):
        # TODO: No need to toggle in assign mode?
        if self.mode == "add":
            status = not self.roi_map.is_variation_used(
                self.physician, self.user_input.GetValue()
            )
            self.button_ok.Enable(status)

    def run(self):
        res = self.ShowModal()
        if res == wx.ID_OK:
            try:
                physician_roi = (
                    self.input_roi_name
                    if self.mode == "add"
                    else self.user_input.GetValue()
                )
                variation = (
                    self.user_input.GetValue()
                    if self.mode == "add"
                    else self.input_roi_name
                )
                self.roi_map.add_variations(
                    physician=self.physician,
                    physician_roi=physician_roi,
                    variation=variation,
                )
            except ROIVariationError as e:
                ROIVariationErrorDialog(self.parent, e)
        self.Destroy()


class AssignVariation(AddVariationDlg):
    """
    Assign a new variation for a specified physician and physician roi
    """

    def __init__(self, parent, physician, roi_map, variation):
        AddVariationDlg.__init__(
            self, parent, physician, roi_map, variation, mode="assign"
        )


class AddVariation(AddVariationDlg):
    """
    Add a new variation for a specified physician and physician roi
    """

    def __init__(self, parent, physician, roi_map, physician_roi):
        AddVariationDlg.__init__(
            self, parent, physician, roi_map, physician_roi, mode="add"
        )


class MoveVariationDialog(wx.Dialog):
    """
    Move provided variations to a new physician roi within the same physician map
    """

    def __init__(
        self,
        parent,
        variations,
        physician,
        old_physician_roi,
        choices,
        roi_map,
    ):
        """
        :param parent: GUI parent
        :param variations: the roi variations to be moved
        :type variations: list
        :param physician: the associated physician
        :type physician: str
        :param old_physician_roi: current physician roi of the provided variations
        :type old_physician_roi: str
        :param choices: the allowed physician rois that the variations may be moved to
        :type choices: list
        :param roi_map: roi_map object
        :type roi_map: DatabaseROIs
        """
        wx.Dialog.__init__(self, parent)
        self.variations = variations
        self.physician = physician
        self.old_physician_roi = old_physician_roi
        self.choices = choices
        self.roi_map = roi_map
        self.combo_box = wx.ComboBox(
            self,
            wx.ID_ANY,
            choices=choices,
            style=wx.CB_DROPDOWN | wx.CB_READONLY,
        )
        self.button_ok = wx.Button(self, wx.ID_OK, "OK")
        self.button_cancel = wx.Button(self, wx.ID_CANCEL, "Cancel")

        self.__set_properties()
        self.__do_layout()

        self.run()

    def __set_properties(self):
        self.SetTitle("Move Variations")
        self.combo_box.SetValue(self.choices[0])

    def __do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_buttons = wx.BoxSizer(wx.HORIZONTAL)
        sizer_input = wx.StaticBoxSizer(
            wx.StaticBox(self, wx.ID_ANY, ""), wx.VERTICAL
        )
        label_physician_roi = wx.StaticText(
            self, wx.ID_ANY, "Move to Physician ROI:"
        )
        sizer_input.Add(label_physician_roi, 0, 0, 0)
        sizer_input.Add(self.combo_box, 0, 0, 0)
        sizer_wrapper.Add(
            sizer_input, 0, wx.EXPAND | wx.TOP | wx.LEFT | wx.RIGHT, 5
        )
        sizer_buttons.Add(self.button_ok, 0, wx.ALL | wx.EXPAND, 5)
        sizer_buttons.Add(self.button_cancel, 0, wx.ALL | wx.EXPAND, 5)
        sizer_wrapper.Add(sizer_buttons, 0, wx.ALIGN_RIGHT | wx.ALL, 5)
        self.SetSizer(sizer_wrapper)
        self.Layout()
        self.Fit()
        self.Center()

    def run(self):
        res = self.ShowModal()
        if res == wx.ID_OK:
            self.action()
        self.Destroy()

    def action(self):
        self.roi_map.delete_variations(
            self.physician, self.old_physician_roi, self.variations
        )
        self.roi_map.add_variations(
            self.physician, self.combo_box.GetValue(), self.variations
        )


class AddPhysicianROI(wx.Dialog):
    """
    Add a new physician roi to the specified physician
    """

    def __init__(
        self,
        parent,
        physician,
        roi_map,
        physician_roi=None,
        institutional_mode=False,
    ):
        """
        :param parent: GUI parent
        :param physician: physician of interest
        :type physician: str
        :param roi_map: roi_map object
        :type roi_map: DatabaseROIs
        :param physician_roi: initial physician roi value
        :type physician_roi: str
        :param institutional_mode: If True, this dialog adds an institutional roi instead
        :type institutional_mode: bool
        """
        wx.Dialog.__init__(self, parent)
        self.SetSize((500, 135))

        self.institutional_mode = institutional_mode

        self.physician = physician
        self.roi_map = roi_map
        self.physician_roi = physician_roi
        self.institutional_rois = self.roi_map.get_unused_institutional_rois(
            physician
        )
        self.text_ctrl_physician_roi = wx.TextCtrl(self, wx.ID_ANY, "")

        if not self.institutional_mode:
            self.combo_box_institutional_roi = wx.ComboBox(
                self,
                wx.ID_ANY,
                choices=self.institutional_rois,
                style=wx.CB_DROPDOWN | wx.CB_READONLY,
            )
        self.button_ok = wx.Button(self, wx.ID_OK, "Add")
        self.button_cancel = wx.Button(self, wx.ID_CANCEL, "Cancel")

        self.res = None

        self.__do_bind()
        self.__set_properties()
        self.__do_layout()

        self.run()

    def __do_bind(self):
        self.Bind(
            wx.EVT_TEXT,
            self.update_enable,
            id=self.text_ctrl_physician_roi.GetId(),
        )

    def __set_properties(self):
        if not self.institutional_mode:
            self.SetTitle("Add Physician ROI for %s" % self.physician)
            self.text_ctrl_physician_roi.SetToolTip(
                "New entry must be unique from all other institutional, physician, "
                "or variation ROIs for this physician."
            )
            self.combo_box_institutional_roi.SetToolTip(
                "If the institutional ROI you’re looking for isn’t here, it may"
                " already be assigned."
            )
            if "uncategorized" in self.institutional_rois:
                self.combo_box_institutional_roi.SetValue("uncategorized")

            self.button_ok.SetToolTip(
                "If Add is disabled, new entry is already used in institutional ROIs, "
                "physician ROIs, or ROI variations for %s." % self.physician
            )
        else:
            self.SetTitle("Add Institutional ROI")
            self.text_ctrl_physician_roi.SetToolTip(
                "New entry must be unique from all other institutional ROIs."
            )
            self.button_ok.SetToolTip(
                "If Add is disabled, new entry is already used in institutional ROIs"
            )

        if self.physician_roi:
            self.text_ctrl_physician_roi.SetValue(self.physician_roi)

        self.SetMinSize((300, 100))

    def __do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_main = wx.BoxSizer(wx.VERTICAL)
        sizer_buttons = wx.BoxSizer(wx.HORIZONTAL)
        sizer_input = wx.StaticBoxSizer(
            wx.StaticBox(self, wx.ID_ANY, ""), wx.HORIZONTAL
        )

        sizer_physician_roi = wx.BoxSizer(wx.VERTICAL)

        label_physician_roi = wx.StaticText(
            self,
            wx.ID_ANY,
            "New %s ROI:"
            % ["Physician", "Institutional"][self.institutional_mode],
        )
        sizer_physician_roi.Add(label_physician_roi, 0, 0, 0)
        sizer_physician_roi.Add(self.text_ctrl_physician_roi, 0, wx.EXPAND, 0)
        sizer_input.Add(sizer_physician_roi, 1, wx.ALL | wx.EXPAND, 5)

        if not self.institutional_mode:
            sizer_institutional_roi = wx.BoxSizer(wx.VERTICAL)
            label_institutional_roi = wx.StaticText(
                self, wx.ID_ANY, "Linked Institutional ROI:"
            )
            sizer_institutional_roi.Add(label_institutional_roi, 0, 0, 0)
            sizer_institutional_roi.Add(
                self.combo_box_institutional_roi, 0, wx.EXPAND, 0
            )
            sizer_input.Add(sizer_institutional_roi, 1, wx.ALL | wx.EXPAND, 5)
        sizer_main.Add(sizer_input, 0, wx.EXPAND, 0)

        sizer_buttons.Add(self.button_ok, 1, wx.ALL, 5)
        sizer_buttons.Add(self.button_cancel, 1, wx.ALL, 5)
        sizer_main.Add(sizer_buttons, 0, wx.ALIGN_RIGHT, 0)
        sizer_wrapper.Add(sizer_main, 1, wx.ALL | wx.EXPAND, 5)

        self.SetSizer(sizer_wrapper)
        self.Layout()
        self.Fit()
        self.Center()

    def run(self):
        self.res = self.ShowModal()
        if self.res == wx.ID_OK:
            self.action()
        self.Destroy()

    def action(self):
        if not self.institutional_mode:
            self.roi_map.add_physician_roi(
                self.physician,
                self.combo_box_institutional_roi.GetValue(),
                self.text_ctrl_physician_roi.GetValue(),
            )

            # in case user changed the name in this dlg
            if self.physician_roi:
                self.roi_map.add_variations(
                    self.physician,
                    self.text_ctrl_physician_roi.GetValue(),
                    self.physician_roi,
                )
        else:
            self.roi_map.add_institutional_roi(
                self.text_ctrl_physician_roi.GetValue()
            )

    def update_enable(self, evt):
        if not self.institutional_mode:
            invalid_choices = set(
                self.roi_map.get_physician_rois(self.physician)
                + self.roi_map.get_all_variations_of_physician(self.physician)
            )
        else:
            invalid_choices = self.roi_map.institutional_rois

        new = clean_name(self.text_ctrl_physician_roi.GetValue())
        self.button_ok.Enable(new not in invalid_choices)


class DelPhysicianROI:
    """Delete a physician roi of the specified physician"""

    def __init__(self, parent, physician, roi_map, physician_roi):
        """
        :param parent: wx parent object
        :param physician: physician associated with physician_roi
        :type physician: str
        :param physician_roi: physician_roi to be deleted
        :type physician_roi: str
        :param roi_map: roi_map to be edited
        """
        caption = "Delete %s (Physician ROI) from %s's ROI map?" % (
            physician_roi,
            physician,
        )
        action = partial(
            roi_map.delete_physician_roi, physician, physician_roi
        )
        MessageDialog(parent, caption, action_yes_func=action)


class DelVariation:
    """Delete a physician roi variation of the specified physician"""

    def __init__(self, parent, physician, roi_map, variation):
        """
        :param parent: wx parent object
        :param physician: physician associated with physician_roi
        :type physician: str
        :param roi_map: roi_map to be edited
        :param variation: variation to be deleted
        :type variation: str
        """
        physician_roi = roi_map.get_physician_roi(physician, variation)
        caption = "Delete %s (Variation) from %s's %s (Physician ROI)?" % (
            variation,
            physician,
            physician_roi,
        )
        action = partial(
            roi_map.delete_variations, physician, physician_roi, variation
        )
        MessageDialog(parent, caption, action_yes_func=action)


class AddROIType(wx.Dialog):
    """
    Add a new roi type (e.g., External, PTV, ORGAN, etc.)
    """

    def __init__(self, parent):
        """
        :param parent: GUI parent
        """
        wx.Dialog.__init__(self, parent)
        self.SetSize((250, 135))
        self.text_ctrl_roi_type = wx.TextCtrl(self, wx.ID_ANY, "")
        self.button_ok = wx.Button(self, wx.ID_OK, "Add")
        self.button_cancel = wx.Button(self, wx.ID_CANCEL, "Cancel")

        self.__set_properties()
        self.__do_layout()

        self.run()

    def __set_properties(self):
        self.SetTitle("Add New ROI Type")

    def __do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_main = wx.BoxSizer(wx.VERTICAL)
        sizer_buttons = wx.BoxSizer(wx.HORIZONTAL)
        sizer_input = wx.StaticBoxSizer(
            wx.StaticBox(self, wx.ID_ANY, ""), wx.HORIZONTAL
        )
        sizer_physician_roi = wx.BoxSizer(wx.VERTICAL)
        label_roi_type = wx.StaticText(self, wx.ID_ANY, "New ROI Type:")
        sizer_physician_roi.Add(label_roi_type, 0, 0, 0)
        sizer_physician_roi.Add(self.text_ctrl_roi_type, 0, wx.EXPAND, 0)
        sizer_input.Add(sizer_physician_roi, 1, wx.ALL | wx.EXPAND, 5)
        sizer_main.Add(sizer_input, 0, wx.EXPAND, 0)
        sizer_buttons.Add(self.button_ok, 1, wx.ALL, 5)
        sizer_buttons.Add(self.button_cancel, 1, wx.ALL, 5)
        sizer_main.Add(sizer_buttons, 0, wx.ALIGN_CENTER, 0)
        sizer_wrapper.Add(sizer_main, 1, wx.ALL | wx.EXPAND, 5)
        self.SetSizer(sizer_wrapper)
        self.Layout()
        self.Center()

    def run(self):
        res = self.ShowModal()
        if res == wx.ID_OK:
            self.action()
        self.Destroy()

    def action(self):
        pass


class ChangePlanROIName(wx.Dialog):
    """
    Change the roi name of a parsed plan in the DICOM importer
    """

    def __init__(
        self,
        tree_ctrl_roi,
        tree_item,
        plan_uid,
        parsed_dicom_data,
        dicom_importer,
    ):
        """
        :param tree_ctrl_roi: the roi tree from the DICOM importer view
        :param tree_item: the tree_ctrl_roi item to be edited
        :param plan_uid: the study instance uid of the plan of interest
        :type plan_uid: str
        :param parsed_dicom_data: parsed dicom data object for plan of interest
        :type parsed_dicom_data: DICOM_Parser
        :param dicom_importer: Needed to edit tree
        :type dicom_importer: DicomTreeBuilder
        """
        wx.Dialog.__init__(
            self, None, title="Edit %s" % tree_ctrl_roi.GetItemText(tree_item)
        )

        self.tree_ctrl_roi = tree_ctrl_roi
        self.tree_item = tree_item
        self.roi = tree_ctrl_roi.GetItemText(tree_item)

        self.plan_uid = plan_uid

        self.parsed_dicom_data = parsed_dicom_data
        self.dicom_importer = dicom_importer

        invalid_options = [""] + parsed_dicom_data.roi_names
        self.invalid_options = [clean_name(name) for name in invalid_options]

        self.text_input_label = "Change ROI name to:"

        self.text_ctrl = wx.TextCtrl(self, wx.ID_ANY, "")
        self.button_ok = wx.Button(self, wx.ID_OK, "OK")
        self.button_cancel = wx.Button(self, wx.ID_CANCEL, "Cancel")

        self.__set_properties()
        self.__do_bind()
        self.__do_layout()

        self.run()

    def __set_properties(self):
        self.text_ctrl.SetMinSize((365, 22))
        self.button_ok.Disable()

    def __do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_ok_cancel = wx.BoxSizer(wx.HORIZONTAL)
        sizer_input = wx.StaticBoxSizer(
            wx.StaticBox(self, wx.ID_ANY, ""), wx.VERTICAL
        )

        label_text_input = wx.StaticText(
            self, wx.ID_ANY, self.text_input_label
        )
        sizer_input.Add(label_text_input, 0, wx.EXPAND | wx.ALL, 5)
        sizer_input.Add(
            self.text_ctrl, 0, wx.BOTTOM | wx.EXPAND | wx.LEFT | wx.RIGHT, 5
        )

        sizer_wrapper.Add(sizer_input, 0, wx.LEFT | wx.RIGHT | wx.TOP, 10)
        sizer_ok_cancel.Add(self.button_ok, 0, wx.ALL, 5)
        sizer_ok_cancel.Add(self.button_cancel, 0, wx.ALL, 5)
        sizer_wrapper.Add(
            sizer_ok_cancel, 0, wx.ALIGN_RIGHT | wx.BOTTOM | wx.RIGHT, 10
        )

        self.SetSizer(sizer_wrapper)
        sizer_wrapper.Fit(self)
        self.Layout()
        self.Center()

    def __do_bind(self):
        self.Bind(wx.EVT_TEXT, self.text_ticker, id=self.text_ctrl.GetId())

    def text_ticker(self, evt):
        [self.button_ok.Disable, self.button_ok.Enable][
            self.new_name not in self.invalid_options
        ]()

    @property
    def roi_key(self):
        return self.parsed_dicom_data.get_roi_key(self.roi)

    @property
    def new_name(self):
        return self.text_ctrl.GetValue()

    def run(self):
        res = self.ShowModal()
        if res == wx.ID_OK:
            self.action()
        self.Destroy()

    def action(self):
        key = self.parsed_dicom_data.get_roi_key(self.roi)
        self.parsed_dicom_data.set_roi_name(key, self.new_name)
        self.tree_ctrl_roi.SetItemText(self.tree_item, self.new_name)
        self.dicom_importer.set_roi_name_over_ride(
            self.plan_uid, self.roi, self.new_name
        )


class RenamerBaseClass(wx.Dialog):
    """
    Simple base class used for renaming (e.g., renaming a physician in the roi map)
    """

    def __init__(
        self, title, text_input_label, invalid_options, is_physician=False
    ):
        """
        :param title: title of the dialog window
        :type title: str
        :param text_input_label: label to be displayed above text_ctrl
        :type text_input_label: str
        :param invalid_options: if text_ctrl is in invalid options, OK button will be disabled
        :type invalid_options: list
        :param is_physician: if editing a physician, clean_name() needs to know
        :type is_physician: bool
        """
        wx.Dialog.__init__(self, None, title=title)

        self.invalid_options = [
            clean_name(option, physician=is_physician)
            for option in invalid_options
        ]
        self.is_physician = is_physician

        self.text_input_label = text_input_label

        self.text_ctrl = wx.TextCtrl(self, wx.ID_ANY, "")
        self.button_ok = wx.Button(self, wx.ID_OK, "OK")
        self.button_cancel = wx.Button(self, wx.ID_CANCEL, "Cancel")

        self.__set_properties()
        self.__do_bind()
        self.__do_layout()

        self.res = None

        self.run()

    def __set_properties(self):
        self.text_ctrl.SetMinSize((365, 22))
        self.button_ok.Disable()

    def __do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_ok_cancel = wx.BoxSizer(wx.HORIZONTAL)
        sizer_input = wx.StaticBoxSizer(
            wx.StaticBox(self, wx.ID_ANY, ""), wx.VERTICAL
        )

        label_text_input = wx.StaticText(
            self, wx.ID_ANY, self.text_input_label
        )
        sizer_input.Add(label_text_input, 0, wx.EXPAND | wx.ALL, 5)
        sizer_input.Add(
            self.text_ctrl, 0, wx.BOTTOM | wx.EXPAND | wx.LEFT | wx.RIGHT, 5
        )

        sizer_wrapper.Add(sizer_input, 0, wx.LEFT | wx.RIGHT | wx.TOP, 10)
        sizer_ok_cancel.Add(self.button_ok, 0, wx.ALL, 5)
        sizer_ok_cancel.Add(self.button_cancel, 0, wx.ALL, 5)
        sizer_wrapper.Add(
            sizer_ok_cancel, 0, wx.ALIGN_RIGHT | wx.BOTTOM | wx.RIGHT, 10
        )

        self.SetSizer(sizer_wrapper)
        sizer_wrapper.Fit(self)
        self.Layout()
        self.Center()

    def __do_bind(self):
        self.Bind(wx.EVT_TEXT, self.text_ticker, id=self.text_ctrl.GetId())

    def text_ticker(self, evt):
        value = clean_name(self.new_name, physician=self.is_physician)
        [self.button_ok.Disable, self.button_ok.Enable][
            value not in self.invalid_options
        ]()

    @property
    def new_name(self):
        return self.text_ctrl.GetValue()

    def run(self):
        self.res = self.ShowModal()
        if self.res == wx.ID_OK:
            self.action()
        self.Destroy()

    def action(self):
        pass


class RenamePhysicianDialog(RenamerBaseClass):
    """
    Dialog for renaming a physician in the roi map
    """

    def __init__(self, physician, roi_map):
        """
        :param physician: physician to be renamed
        :type physician: str
        :param roi_map: roi_map object
        :type roi_map: DatabaseROIs
        """
        self.physician = physician
        self.roi_map = roi_map
        RenamerBaseClass.__init__(
            self,
            "Rename %s" % physician,
            "New Physician Name:",
            list(roi_map.physicians),
            is_physician=True,
        )

    def action(self):
        self.roi_map.rename_physician(self.new_name, self.physician)


class RenamePhysicianROIDialog(RenamerBaseClass):
    """
    Dialog to rename a physician roi for a given physician
    """

    def __init__(self, physician, physician_roi, roi_map):
        """
        :param physician: physician of the physician roi to be renamed
        :type physician: str
        :param physician_roi: physician roi to be renamed
        :type physician_roi: str
        :param roi_map: roi_map object
        :type roi_map: DatabaseROIs
        """
        self.physician = physician
        self.physician_roi = physician_roi
        self.roi_map = roi_map
        RenamerBaseClass.__init__(
            self,
            "Rename %s for %s" % (physician_roi, physician),
            "New Physician ROI name:",
            roi_map.get_physician_rois(physician),
        )

    def action(self):
        self.roi_map.rename_physician_roi(
            self.new_name, self.physician, self.physician_roi
        )


class RenameInstitutionalROIDialog(RenamerBaseClass):
    """
    Dialog to rename an institutional roi
    """

    def __init__(self, institutional_roi, roi_map):
        """
        :param institutional_roi: institutional roi to be renamed
        :type institutional_roi: str
        :param roi_map: roi_map object
        :type roi_map: DatabaseROIs
        """
        self.institutional_roi = institutional_roi
        self.roi_map = roi_map
        RenamerBaseClass.__init__(
            self,
            "Rename %s" % institutional_roi,
            "New Institutional ROI name:",
            roi_map.institutional_rois,
        )

    def action(self):
        self.roi_map.rename_institutional_roi(
            self.new_name, self.institutional_roi
        )


class LinkPhysicianROI(wx.Dialog):
    """
    Dialog to change which institutional roi a specified physician roi is linked
    """

    def __init__(self, parent, physician, physician_roi, roi_map):
        """
        :param parent: GUI parent
        :param physician: physician for the specified physician roi
        :type physician: str
        :param physician_roi: physician roi of interest
        :type physician_roi: str
        :param roi_map: roi_map object
        :type roi_map: DatabaseROIs
        """
        wx.Dialog.__init__(self, parent)
        self.SetMinSize((500, 135))

        self.physician = physician
        self.physician_roi = physician_roi
        self.roi_map = roi_map
        self.institutional_rois = self.roi_map.institutional_rois
        self.institutional_roi = self.roi_map.get_institutional_roi(
            self.physician, self.physician_roi
        )
        choices = self.roi_map.get_unused_institutional_rois(self.physician)
        if self.institutional_roi not in choices:
            choices.append(self.institutional_roi)
            choices.sort()
        self.combo_box_institutional_roi = wx.ComboBox(
            self,
            wx.ID_ANY,
            choices=choices,
            style=wx.CB_DROPDOWN | wx.CB_READONLY,
        )
        self.button_ok = wx.Button(self, wx.ID_OK, "OK")
        self.button_cancel = wx.Button(self, wx.ID_CANCEL, "Cancel")

        self.res = None

        self.__set_properties()
        self.__do_layout()

        self.run()

    def __set_properties(self):
        self.SetTitle("Change Institutional ROI for %s" % self.physician_roi)
        tip = "If the institutional ROI you’re looking for isn’t here, it may already be assigned."
        self.combo_box_institutional_roi.SetToolTip(tip)
        self.button_ok.SetToolTip(tip)
        if "uncategorized" in self.institutional_rois:
            self.combo_box_institutional_roi.SetValue("uncategorized")

        self.combo_box_institutional_roi.SetValue(self.institutional_roi)

    def __do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_main = wx.BoxSizer(wx.VERTICAL)
        sizer_buttons = wx.BoxSizer(wx.HORIZONTAL)
        sizer_input = wx.StaticBoxSizer(
            wx.StaticBox(self, wx.ID_ANY, ""), wx.HORIZONTAL
        )
        sizer_institutional_roi = wx.BoxSizer(wx.VERTICAL)
        sizer_physician_roi = wx.BoxSizer(wx.VERTICAL)

        sizer_input.Add(sizer_physician_roi, 1, wx.ALL | wx.EXPAND, 5)

        label_institutional_roi = wx.StaticText(
            self, wx.ID_ANY, "Linked Institutional ROI:"
        )
        sizer_institutional_roi.Add(label_institutional_roi, 0, 0, 0)
        sizer_institutional_roi.Add(
            self.combo_box_institutional_roi, 0, wx.EXPAND, 0
        )
        sizer_input.Add(sizer_institutional_roi, 1, wx.ALL | wx.EXPAND, 5)
        sizer_main.Add(sizer_input, 0, wx.EXPAND, 0)

        sizer_buttons.Add(self.button_ok, 1, wx.ALL, 5)
        sizer_buttons.Add(self.button_cancel, 1, wx.ALL, 5)
        sizer_main.Add(sizer_buttons, 0, wx.ALIGN_RIGHT, 0)
        sizer_wrapper.Add(sizer_main, 1, wx.ALL | wx.EXPAND, 5)

        self.SetSizer(sizer_wrapper)
        self.Layout()
        self.Fit()
        self.Center()

    def run(self):
        self.res = self.ShowModal()
        if self.res == wx.ID_OK:
            self.action()
        self.Destroy()

    def action(self):
        self.roi_map.set_linked_institutional_roi(
            self.combo_box_institutional_roi.GetValue(),
            self.physician,
            self.physician_roi,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# tools.stats.py
"""
Take numerical data from main app and convert to a format suitable for statistical analysis
in Regression and Control Chart tabs

"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

import numpy as np
from scipy import stats as scipy_stats
from sklearn import linear_model
from sklearn.metrics import mean_squared_error, r2_score
from regressors import stats as regressors_stats
from dvha.db import sql_columns


class StatsData:
    """Class used to to collect data for Regression and Control Chart
    This process is different than for Time Series since regressions require
    all variables to be the same length

    Parameters
    ----------
    dvhs : DVH
        data from DVH query
    table_data : dict
        table data other than from DVHs. Has keys of 'Plans', 'Rxs', 'Beams'
        with values of QuerySQL objects

    """

    def __init__(self, dvhs, table_data, group=1):

        self.dvhs = dvhs
        self.table_data = table_data
        self.group = group

        self.column_info = sql_columns.numerical
        self.correlation_variables = list(self.column_info)
        self.correlation_variables.sort()

        self.__map_data()

    def __map_data(self):
        self.data = {}
        stat_types = ["min", "mean", "median", "max"]
        for var in self.correlation_variables:
            if var in self.column_info.keys():
                var_name = self.column_info[var]["var_name"]
                table = self.column_info[var]["table"]

                if table == "DVHs" and hasattr(self.dvhs, var_name):
                    self.data[var] = {
                        "units": self.column_info[var]["units"],
                        "values": getattr(self.dvhs, var_name),
                    }

                # single value variables
                elif table == "Plans" and hasattr(
                    self.table_data[table], var_name
                ):
                    src = self.table_data[table]
                    self.data[var] = {
                        "units": self.column_info[var]["units"],
                        "values": [
                            getattr(src, var_name)[self.get_plan_index(uid)]
                            for uid in self.uids
                        ],
                    }

                # multi value variables
                elif table == "Beams" and hasattr(
                    self.table_data[table], var_name
                ):
                    src = self.table_data[table]

                    if str_starts_with_any_in_list(
                        var,
                        [
                            "Beam Complexity",
                            "Beam Area",
                            "Control Point MU",
                            "Beam Perimeter",
                            "Beam Energy",
                        ],
                    ):
                        # stats of these four variable types have min, mean, median, and max types in DB
                        # The following will take min, mean, median, or max of all values for a UID based on var type
                        # Example, if var_name == Beam Complexity (Max), the following will return the Max of these
                        temp = []
                        for uid in self.uids:
                            indices = self.get_beam_indices(uid)
                            beam_data = getattr(
                                self.table_data["Beams"], var_name
                            )
                            values = [
                                beam_data[i]
                                for i in indices
                                if beam_data[i] != "None"
                            ]
                            for stat in stat_types:
                                if stat in var.lower():
                                    if values:
                                        temp.append(getattr(np, stat)(values))
                                    else:
                                        temp.append(None)
                        self.data[var] = {
                            "units": self.column_info[var]["units"],
                            "values": temp,
                        }
                    else:
                        temp = {s: [] for s in stat_types}
                        for uid in self.uids:
                            for stat in stat_types:
                                values = self._get_src_values(
                                    src, var_name, uid
                                )
                                values = [v for v in values if v != "None"]
                                if values:
                                    temp[stat].append(
                                        getattr(np, stat)(values)
                                    )
                                else:
                                    temp[stat].append(None)

                        for stat in stat_types:
                            corr_key = "%s (%s)" % (var, stat.capitalize())
                            self.data[corr_key] = {
                                "units": self.column_info[var]["units"],
                                "values": temp[stat],
                            }
        self._validate_data()

    def _validate_data(self):
        """Remove any variables that are constant to avoid crash on
        regression"""
        bad_vars = []
        for var_name, var_obj in self.data.items():
            if "Date" in var_name:
                if var_name != "Simulation Date":
                    bad_vars.append(var_name)
            else:
                values = [
                    float(val)
                    for val in var_obj["values"]
                    if val != "None" and val is not None
                ]
                if not any(np.diff(values).tolist()):
                    bad_vars.append(var_name)

        for var in bad_vars:
            self.data.pop(var)

    def update_endpoints_and_radbio(self):
        """Update endpoint and radbio data in self.data. This function is
        needed since all of these values are calculated
        after a query and user may change these values.
        """
        if self.dvhs:
            if self.dvhs.endpoints["defs"]:
                for var in self.dvhs.endpoints["defs"]["label"]:
                    if var not in self.variables:
                        self.data[var] = {
                            "units": "",
                            "values": self.dvhs.endpoints["data"][var],
                        }

                for var in self.variables:
                    if var[0:2] in {"D_", "V_"}:
                        if var not in self.dvhs.endpoints["defs"]["label"]:
                            self.data.pop(var)

            if self.dvhs.eud:
                self.data["EUD"] = {"units": "Gy", "values": self.dvhs.eud}
            if self.dvhs.ntcp_or_tcp:
                self.data["NTCP or TCP"] = {
                    "units": "",
                    "values": self.dvhs.ntcp_or_tcp,
                }
            self._validate_data()

    @staticmethod
    def _get_src_values(src, var_name, uid):
        """

        Parameters
        ----------
        src : QuerySQL
            table data
        var_name : str
            attribute to be extracted from table data
        uid : str
            StudyInstanceUID stored in the SQL database

        Returns
        -------
        list
            values of ``var_name`` from table data that match ``uid``
        """
        uid_indices = [
            i for i, x in enumerate(src.study_instance_uid) if x == uid
        ]
        return [getattr(src, var_name)[i] for i in uid_indices]

    def get_plan_index(self, uid):
        """Get the index of ``uid`` from the Plans table

        Parameters
        ----------
        uid : str
            StudyInstanceUID as stored in the SQL database

        Returns
        -------
        int
            Plans table index for ``uid``

        """
        return self.table_data["Plans"].study_instance_uid.index(uid)

    def get_beam_indices(self, uid):
        """Get the indices of the Beams table with ``uid``

        Parameters
        ----------
        uid : str
            StudyInstanceUID as stored in the SQL database

        Returns
        -------
        list
            Beams table indices that match ``uid``

        """
        return [
            i
            for i, x in enumerate(self.table_data["Beams"].study_instance_uid)
            if x == uid
        ]

    def get_bokeh_data(self, x, y):
        """Get data in a format compatible with bokeh's ColumnDataSource.data

        Parameters
        ----------
        x : str
            x-variable name
        y : str
            y-variable name

        Returns
        -------
        dict
            x and y data

        """
        if x in list(self.data) and y in list(self.data):
            # TODO: potential data length issue with study_instance_uid
            # Received the following error with 0.6.9, can't reproduce
            # BokehUserWarning: ColumnDataSource's columns must be of the
            # same length. Current lengths:
            # ('date', 69), ('mrn', 69), ('uid', 70), ('x', 69), ('y', 69)
            # Appears to point to this function's return?
            return {
                "uid": self.uids,
                "mrn": self.mrns,
                "date": self.sim_study_dates,
                "x": self.data[x]["values"],
                "y": self.data[y]["values"],
            }
        return {key: [] for key in ["uid", "mrn", "date", "x", "y"]}

    @property
    def uids(self):
        """StudyInstanceUIDs from DVH object

        Returns
        -------
        list
            ``DVH.study_instance_uid``
        """
        return self.dvhs.study_instance_uid

    @property
    def mrns(self):
        """MRNs from DVH object

        Returns
        -------
        list
            ``DVH.mrn``
        """
        return self.dvhs.mrn

    @property
    def sim_study_dates(self):
        """Simulation dates from Plans table

        Returns
        -------
        list
            Simulation dates
        """
        return self.data["Simulation Date"]["values"]

    @property
    def variables(self):
        """Get variable names for plotting

        Returns
        -------
        list
            keys of ``StatsData.data`` sans 'Simulation Date'

        """
        return [var for var in list(self.data) if var != "Simulation Date"]

    @property
    def vars_with_nan_values(self):
        """Find variable names that contain non-numerical values

        Returns
        -------
        list
            Variable names that cannot be converted to ``float``
        """
        ans = []
        for var in self.variables:
            for val in self.data[var]["values"]:
                try:
                    float(val)
                except Exception:
                    ans.append(var)
                    break
        return ans

    def get_axis_title(self, variable):
        """Get the plot axis title for ``variable``

        Parameters
        ----------
        variable : str
            A key of ``StatsData.data``

        Returns
        -------
        str
            ``variable`` with units if stored

        """
        if self.data[variable]["units"]:
            return "%s (%s)" % (variable, self.data[variable]["units"])
        return variable

    def get_X_and_y(self, y_variable, x_variables, include_patient_info=False):
        """Collect data for input into multi-variable regression

        Parameters
        ----------
        y_variable : str
            dependent variable
        x_variables : list
            independent variables
        include_patient_info : bool
            If True, return mrn, uid, dates with X and y

        Returns
        -------
        type
            X, y or X, y, mrn, uid, dates

        """
        data, mrn, uid, dates = [], [], [], []
        y_var_data = []
        for i, value in enumerate(self.data[y_variable]["values"]):
            y_var_data.append([value, np.nan][value == "None"])
            mrn.append(self.mrns[i])
            uid.append(self.uids[i])
            dates.append(self.sim_study_dates[i])

        data.append(y_var_data)
        for var in x_variables:
            x_var_data = []
            for value in self.data[var]["values"]:
                x_var_data.append(
                    [value, np.nan][str(value).lower() == "none"]
                )
            data.append(x_var_data)

        data = np.array(data)
        bad_indices = get_index_of_nan(data)

        for bad_index in bad_indices[::-1]:
            data = np.delete(data, bad_index, 1)
            mrn.pop(bad_index)
            uid.pop(bad_index)
            dates.pop(bad_index)

        X = np.transpose(data[1:])
        y = data[0]

        if not include_patient_info:
            return X, y
        return X, y, mrn, uid, dates

    def add_variable(self, variable, values, units=""):
        """Add a new variable to ``StatsData.data``, will not over-write

        Parameters
        ----------
        variable : str
            variable name to be used as a key and plot title
        values : list
            values to be stored for ``variable``
        units : str, optional
             Define units for display on plot

        """
        if variable not in list(self.data):
            self.data[variable] = {"units": units, "values": values}
        if variable not in self.correlation_variables:
            self.correlation_variables.append(variable)
            self.correlation_variables.sort()

    def del_variable(self, variable):
        """Delete a variable from ``StatsData.data``

        Parameters
        ----------
        variable : str
            variable name

        """
        if variable in list(self.data):
            self.data.pop(variable)
        if variable in self.correlation_variables:
            index = self.correlation_variables.index(variable)
            self.correlation_variables.pop(index)

    def set_variable_data(self, variable, data, units=None):
        """Replace the data for the given variable in ``StatsData.data``

        Parameters
        ----------
        variable : str
            variable name
        data : list
            new data
        units : str, optional
             Define units for display on plot

        """
        self.data[variable]["values"] = data
        if units is not None:
            self.data[variable]["units"] = units

    def set_variable_units(self, variable, units):
        """Set the units for the given variable in ``StatsData.data``

        Parameters
        ----------
        variable : str
            variable name
        units : str
            units for display on plot

        """
        self.data[variable]["units"] = units

    def get_corr_matrix_data(
        self, options, included_vars=None, extra_vars=None
    ):
        """Get a Pearson-R correlation matrix

        Parameters
        ----------
        options : dvha.options.Options
            DVHA options class object. Used to get colors.
        included_vars : list, optional
             variables to be included in matrix
        extra_vars : list, optional
             variables to be excluded from the matrix

        Returns
        -------
        tuple (dict, list)
            The dictionary has keys of 'source_data', 'x_factors' and
            'y_factors'. `source_data` is used for bokeh plotting, the factors
            are for axis tick labels. The 2nd parameter of the tuple is a list
            of removed mrns

        """
        if included_vars is None:
            included_vars = list(self.data)
        if extra_vars is not None:
            included_vars = included_vars + extra_vars
        else:
            extra_vars = []

        categories = [
            c
            for c in list(self.data)
            if "date" not in c.lower() and c in included_vars
        ]
        categories.extend(extra_vars)
        categories = list(set(categories))
        categories.sort()
        var_count = len(categories)
        categories_for_label = [
            category.replace("Control Point", "CP") for category in categories
        ]
        categories_for_label = [
            category.replace("control point", "CP")
            for category in categories_for_label
        ]
        categories_for_label = [
            category.replace("Distance", "Dist")
            for category in categories_for_label
        ]

        for i, category in enumerate(categories_for_label):
            if category.startswith("DVH"):
                categories_for_label[i] = category.split("DVH Endpoint: ")[1]

        x_factors = categories_for_label
        y_factors = categories_for_label[::-1]

        s_keys = [
            "x",
            "y",
            "x_name",
            "y_name",
            "color",
            "alpha",
            "r",
            "p",
            "size",
            "x_normality",
            "y_normality",
            "group",
        ]
        source_data = {
            "corr": {sk: [] for sk in s_keys},
            "line": {"x": [0.5, var_count - 0.5], "y": [var_count - 0.5, 0.5]},
        }

        min_size, max_size = 3, 20
        removed_mrns = set()
        for x in range(var_count):
            for y in range(var_count):
                if x > y and self.group == 1 or x < y and self.group == 2:
                    if (
                        categories[x] not in extra_vars
                        and categories[y] not in extra_vars
                    ):

                        bad_indices = [
                            i
                            for i, v in enumerate(
                                self.data[categories[x]]["values"]
                            )
                            if type(v) in [str, type(None)]
                        ]
                        bad_indices.extend(
                            [
                                i
                                for i, v in enumerate(
                                    self.data[categories[y]]["values"]
                                )
                                if type(v) in [str, type(None)]
                            ]
                        )
                        bad_indices = list(set(bad_indices))
                        removed_mrns = removed_mrns.union(
                            set(self.mrns[i] for i in bad_indices)
                        )

                        x_data = [
                            v
                            for i, v in enumerate(
                                self.data[categories[x]]["values"]
                            )
                            if i not in bad_indices
                        ]
                        y_data = [
                            v
                            for i, v in enumerate(
                                self.data[categories[y]]["values"]
                            )
                            if i not in bad_indices
                        ]

                        if x_data and len(x_data) == len(y_data):
                            r, p_value = scipy_stats.pearsonr(x_data, y_data)
                        else:
                            r, p_value = 0, 0
                        if np.isnan(r):
                            r = 0

                        sign = ["neg", "pos"][r >= 0]
                        color = getattr(
                            options,
                            "CORRELATION_%s_COLOR_%s"
                            % (sign.upper(), self.group),
                        )
                        source_data["corr"]["color"].append(color)
                        source_data["corr"]["r"].append(r)
                        source_data["corr"]["p"].append(p_value)
                        source_data["corr"]["alpha"].append(abs(r))
                        source_data["corr"]["size"].append(
                            ((max_size - min_size) * abs(r)) + min_size
                        )
                        source_data["corr"]["x"].append(
                            x + 0.5
                        )  # 0.5 offset due to bokeh 0.12.9 bug
                        source_data["corr"]["y"].append(
                            var_count - y - 0.5
                        )  # 0.5 offset due to bokeh 0.12.9 bug
                        source_data["corr"]["x_name"].append(
                            categories_for_label[x]
                        )
                        source_data["corr"]["y_name"].append(
                            categories_for_label[y]
                        )
                        source_data["corr"]["group"].append(self.group)

                        try:
                            x_norm, x_p = scipy_stats.normaltest(x_data)
                        except ValueError:
                            x_p = "N/A"
                        try:
                            y_norm, y_p = scipy_stats.normaltest(y_data)
                        except ValueError:
                            y_p = "N/A"

                        source_data["corr"]["x_normality"].append(x_p)
                        source_data["corr"]["y_normality"].append(y_p)

        return {
            "source_data": source_data,
            "x_factors": x_factors,
            "y_factors": y_factors,
        }, removed_mrns


def get_index_of_nan(numpy_array):
    """Find indices of np.nan values

    Parameters
    ----------
    numpy_array : np.ndarray
        A numpy array

    Returns
    -------
    list
        indices of ``numpy_array`` that are ``np.nan``
    """
    bad_indices = []
    nan_data = np.isnan(numpy_array)
    for var_data in nan_data:
        indices = np.where(var_data)[0].tolist()
        if indices:
            bad_indices.extend(indices)
    bad_indices = list(set(bad_indices))
    bad_indices.sort()
    return bad_indices


def str_starts_with_any_in_list(string_a, string_list):
    """Check if string_a starts with any string the provided list of strings

    Parameters
    ----------
    string_a : str
        Any string
    string_list : list
        A list of strings

    Returns
    -------
    bool
        True if any ``string_a`` starts with any string in ``string_list``

    """
    for string_b in string_list:
        if string_a.startswith(string_b):
            return True
    return False


def get_p_values(X, y, predictions, params):
    """Get p-values using sklearn
    based on https://stackoverflow.com/questions/27928275/find-p-value-significance-in-scikit-learn-linearregression

    Parameters
    ----------
    X : np.ndarray
        independent data
    y : np.ndarray
        dependent data
    predictions :
        output from linear_model.LinearRegression.predict
    params :
        np.array([y_incercept, slope])

    Returns
    -------
    list
        p-values

    """

    newX = np.append(np.ones((len(X), 1)), X, axis=1)
    MSE = (sum((y - predictions) ** 2)) / (len(newX) - len(newX[0]))

    var_b = MSE * (np.linalg.inv(np.dot(newX.T, newX)).diagonal())
    sd_b = np.sqrt(var_b)
    ts_b = params / sd_b

    return (
        [
            2 * (1 - scipy_stats.t.cdf(np.abs(i), (len(newX) - 1)))
            for i in ts_b
        ],
        sd_b,
        ts_b,
    )


class MultiVariableRegression:
    """Perform a multi-variable regression using sklearn

    Parameters
    ----------
    X : np.array
        independent data
    y : list
        dependent data

    """

    def __init__(self, X, y, saved_reg=None):

        if saved_reg is None:
            self.reg = linear_model.LinearRegression()
            self.ols = self.reg.fit(X, y)
        else:
            self.reg = saved_reg.reg
            self.ols = saved_reg.ols

        self.y_intercept = self.reg.intercept_
        self.slope = self.reg.coef_
        params = np.append(self.y_intercept, self.slope)
        self.predictions = self.reg.predict(X)

        self.r_sq = r2_score(y, self.predictions)
        self.mse = mean_squared_error(y, self.predictions)

        self.p_values, self.sd_b, self.ts_b = get_p_values(
            X, y, self.predictions, params
        )

        self.residuals = np.subtract(y, self.predictions)

        self.norm_prob_plot = scipy_stats.probplot(
            self.residuals, dist="norm", fit=False, plot=None, rvalue=False
        )

        reg_prob = linear_model.LinearRegression()
        reg_prob.fit(
            [[val] for val in self.norm_prob_plot[0]], self.norm_prob_plot[1]
        )

        self.y_intercept_prob = reg_prob.intercept_
        self.slope_prob = reg_prob.coef_
        self.x_trend_prob = [
            min(self.norm_prob_plot[0]),
            max(self.norm_prob_plot[0]),
        ]
        self.y_trend_prob = np.add(
            np.multiply(self.x_trend_prob, self.slope_prob),
            self.y_intercept_prob,
        )

        self.f_stat = regressors_stats.f_stat(self.ols, X, y)
        self.df_error = len(X[:, 0]) - len(X[0, :]) - 1
        self.df_model = len(X[0, :])

        self.f_p_value = scipy_stats.f.cdf(
            self.f_stat, self.df_model, self.df_error
        )


def get_control_limits(y, std_devs=3):
    """Calculate control limits for Control Chart

    Parameters
    ----------
    y : list
        data
    std_devs : int or float
        values greater than std_devs away are out-of-control (Default value = 3)

    Returns
    -------
    type
        center line, upper control limit, and lower control limit

    """
    y = np.array(y)

    center_line = np.mean(y)
    avg_moving_range = np.mean(np.absolute(np.diff(y)))

    scalar_d = (
        1.128  # since moving range is calculated based on 2 consecutive points
    )

    ucl = center_line + std_devs * avg_moving_range / scalar_d
    lcl = center_line - std_devs * avg_moving_range / scalar_d

    return center_line, ucl, lcl


def sync_variables_in_stats_data_objects(stats_data_1, stats_data_2):
    """Ensure both stats_data objects have the same variables

    Parameters
    ----------
    stats_data_1 : StatsData
        A StatsData object (e.g., Group 1)
    stats_data_2 : StatsData
        Another StatsData object (e.g., Group 2)

    """

    stats_data = {1: stats_data_1, 2: stats_data_2}
    variables = {grp: list(sd.data) for grp, sd in stats_data.items()}

    for grp, sd in stats_data.items():
        for var in variables[grp]:
            if var not in variables[3 - grp]:
                values = ["None"] * len(stats_data[3 - grp].mrns)
                units = stats_data[grp].data[var]["units"]
                stats_data[3 - grp].add_variable(var, values, units)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# threading_progress.py
"""
Generic classes to perform threading with a progress frame
"""
# Copyright (c) 2020 Dan Cutright
# This file is part of DVHA DICOM Editor, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVHA-DICOM-Editor

import wx
from pubsub import pub
from queue import Queue
from threading import Thread
from time import sleep


class ProgressFrame(wx.Dialog):
    """Create a window to display progress and begin provided worker"""

    def __init__(
        self,
        obj_list,
        action,
        close_msg=None,
        action_msg=None,
        action_gui_phrase="Processing",
        title="Progress",
        sub_gauge=False,
        kwargs=False,
        star_map=False,
    ):
        wx.Dialog.__init__(self, None)

        self.close_msg = close_msg
        self.worker_args = [
            obj_list,
            action,
            action_msg,
            action_gui_phrase,
            title,
            kwargs,
            star_map,
        ]
        self.action_gui_phrase = action_gui_phrase

        self.gauge = wx.Gauge(self, wx.ID_ANY, 100)
        self.sub_gauge = wx.Gauge(self, wx.ID_ANY, 100) if sub_gauge else None
        self.label = wx.StaticText(self, wx.ID_ANY, "Initializing...")
        self.sub_label = (
            wx.StaticText(self, wx.ID_ANY, "") if sub_gauge else None
        )

        self.__set_properties()
        self.__do_subscribe()
        self.__do_layout()

        self.run()

    def run(self):
        """Initiate layout in GUI and begin thread"""
        self.Show()
        ProgressFrameWorker(*self.worker_args)

    def __set_properties(self):
        self.SetMinSize((672, 100))

    def __do_subscribe(self):
        pub.subscribe(self.update, "progress_update")
        pub.subscribe(self.sub_update, "sub_progress_update")
        pub.subscribe(self.set_title, "progress_set_title")
        pub.subscribe(self.close, "progress_close")

    @staticmethod
    def __do_unsubscribe():
        pub.unsubAll(topicName="progress_update")
        pub.unsubAll(topicName="sub_progress_update")
        pub.unsubAll(topicName="progress_set_title")
        pub.unsubAll(topicName="progress_close")

    def __do_layout(self):
        sizer_wrapper = wx.BoxSizer(wx.VERTICAL)
        sizer_objects = wx.BoxSizer(wx.VERTICAL)
        sizer_objects.Add(self.label, 0, 0, 0)
        sizer_objects.Add(self.gauge, 0, wx.EXPAND, 0)
        if self.sub_gauge is not None:
            sizer_objects.Add(self.sub_label, 0, 0, 0)
            sizer_objects.Add(self.sub_gauge, 0, wx.EXPAND, 0)
        sizer_wrapper.Add(sizer_objects, 0, wx.ALL | wx.EXPAND, 10)
        self.SetSizer(sizer_wrapper)
        self.Fit()
        self.Layout()
        self.Center()

    def set_title(self, msg):
        """

        Parameters
        ----------
        msg :


        Returns
        -------

        """
        wx.CallAfter(self.SetTitle, msg)

    def update(self, msg):
        """Update the progress message and gauge

        Parameters
        ----------
        msg : dict
            a dictionary with keys of 'label' and 'gauge' text and progress fraction, respectively

        Returns
        -------

        """
        label = msg["label"]
        wx.CallAfter(self.label.SetLabelText, label)
        wx.CallAfter(self.gauge.SetValue, int(100 * msg["gauge"]))

    def sub_update(self, msg):
        """

        Parameters
        ----------
        msg :


        Returns
        -------

        """
        label = msg["label"]
        wx.CallAfter(self.sub_label.SetLabelText, label)
        wx.CallAfter(self.sub_gauge.SetValue, int(100 * msg["gauge"]))

    def close(self):
        """Destroy layout in GUI and send message close message for parent"""
        if self.close_msg is not None:
            wx.CallAfter(pub.sendMessage, self.close_msg)
        self.__do_unsubscribe()
        wx.CallAfter(self.Destroy)


class ProgressFrameWorker(Thread):
    """Create a thread, perform action on each item in obj_list"""

    def __init__(
        self,
        obj_list,
        action,
        action_msg,
        action_gui_phrase,
        title,
        kwargs,
        star_map,
    ):
        Thread.__init__(self)

        pub.sendMessage("progress_set_title", msg=title)

        self.obj_list = obj_list
        self.obj_count = len(self.obj_list)
        self.action = action
        self.action_msg = action_msg
        self.action_gui_phrase = action_gui_phrase
        self.kwargs = kwargs
        self.star_map = star_map

        self.start()

    def run(self):
        """ """
        queue = self.get_queue()
        worker = Thread(target=self.target, args=[queue])
        worker.setDaemon(True)
        worker.start()
        queue.join()
        sleep(0.3)  # Allow time for user to see final progress in GUI
        pub.sendMessage("progress_close")

    def get_queue(self):
        """ """
        queue = Queue()
        for i, obj in enumerate(self.obj_list):
            msg = {
                "label": "%s (%s of %s)"
                % (self.action_msg, i + 1, len(self.obj_list)),
                "gauge": float(i / len(self.obj_list)),
            }
            queue.put((obj, msg))
        return queue

    def target(self, queue):
        """

        Parameters
        ----------
        queue :


        Returns
        -------

        """
        while queue.qsize():
            parameters = queue.get()
            self.do_action(*parameters)
            queue.task_done()

        msg = {"label": "Process Complete", "gauge": 1.0}
        wx.CallAfter(pub.sendMessage, "progress_update", msg=msg)

    def do_action(self, obj, msg):
        """

        Parameters
        ----------
        obj :

        msg :


        Returns
        -------

        """
        wx.CallAfter(pub.sendMessage, "progress_update", msg=msg)

        if self.kwargs:
            result = self.action(**obj)
        elif self.star_map:
            result = self.action(*obj)
        else:
            result = self.action(obj)
        if self.action_msg is not None:
            msg = {"obj": obj, "data": result}
            wx.CallAfter(pub.sendMessage, self.action_msg, msg=msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# db.sql_connector.py
"""Tools used to communicate with the SQL database"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

from wx import CallAfter
import psycopg2
import sqlite3
from datetime import datetime
from dateutil.parser import parse as date_parser
from os.path import dirname, join, isfile
from dvha.db.sql_columns import categorical, numerical
from dvha.options import Options
from dvha.paths import CREATE_PGSQL_TABLES, CREATE_SQLITE_TABLES, DATA_DIR
from dvha.tools.errors import SQLError, push_to_log
import json


class DVH_SQL:
    """This class is used to communicate to the SQL database

    Parameters
    ----------
    config : dict
        optional SQL login credentials, stored values used if nothing provided.
        Allowed keys are 'host', 'port', 'dbname', 'user', 'password'
    db_type : str, optional
        either 'pgsql' or 'sqlite'
    group : int, optional
        use a group-specific connection, either 1 or 2
    """

    def __init__(self, *config, db_type=None, group=1):

        stored_options = Options()

        if config:
            self.db_type = db_type if db_type is not None else "pgsql"
            config = config[0]
        else:
            # Read SQL configuration file
            if group == 2 and stored_options.SYNC_SQL_CNX:
                group = 1
            self.db_type = (
                stored_options.DB_TYPE_GRPS[group]
                if db_type is None
                else db_type
            )
            config = stored_options.SQL_LAST_CNX_GRPS[group][self.db_type]

        if self.db_type == "sqlite":
            db_file_path = config["host"]
            if not dirname(
                db_file_path
            ):  # file_path has not directory, assume it lives in DATA_DIR
                db_file_path = join(DATA_DIR, db_file_path)
            self.db_name = None
            self.cnx = sqlite3.connect(db_file_path)
        else:
            self.db_name = config["dbname"]
            self.cnx = psycopg2.connect(**config)

        self.cursor = self.cnx.cursor()
        self.tables = ["DVHs", "Plans", "Rxs", "Beams", "DICOM_Files"]

    def __enter__(self):
        return self

    def __exit__(self, ctx_type, ctx_value, ctx_traceback):
        self.close()

    def close(self):
        """Close the SQL DB connection"""
        self.cnx.close()

    def execute_file(self, sql_file_name):
        """Executes lines within provided text file to SQL

        Parameters
        ----------
        sql_file_name : str
            absolute file path of a text file containing SQL commands

        """

        for line in open(sql_file_name):
            if not line.startswith("--"):  # ignore commented lines
                self.cursor.execute(line)
        self.cnx.commit()

    def execute_str(self, command_str):
        """Execute and commit a string in proper SQL syntax, can handle
        multiple lines split by \n

        Parameters
        ----------
        command_str : str
            command or commands to be executed and committed

        """
        for line in command_str.split("\n"):
            if line:
                self.cursor.execute(line)
        self.cnx.commit()

    def check_table_exists(self, table_name):
        """Check if a table exists

        Parameters
        ----------
        table_name : st
            the SQL table to check

        Returns
        -------
        bool
            True if ``table_name`` exists

        """

        self.cursor.execute(
            """
            SELECT COUNT(*)
            FROM information_schema.tables
            WHERE table_name = '{0}'
            """.format(
                table_name.replace("'", "''")
            )
        )
        return self.cursor.fetchone()[0] == 1

    def query(self, table_name, return_col_str, *condition_str, **kwargs):
        """A generalized query function for DVHA

        Parameters
        ----------
        table_name : str
            DVHs', 'Plans', 'Rxs', 'Beams', or 'DICOM_Files'
        return_col_str : str: str
            a csv of SQL columns to be returned
        condition_str : str: str
            a condition in SQL syntax
        kwargs :
            optional parameters order, order_by, and bokeh_cds

        Returns
        -------
        list, dict
            Returns a list of lists by default, or a dict of lists if
            bokeh_cds in kwargs and is true
        """
        order, order_by = None, None
        if kwargs:
            if "order" in kwargs:
                order = kwargs["order"]
            if "order_by" in kwargs:
                order_by = kwargs["order_by"]
                if not order:
                    order = "ASC"

        query = "Select %s from %s;" % (return_col_str, table_name)
        if condition_str and condition_str[0]:
            query = "Select %s from %s where %s;" % (
                return_col_str,
                table_name,
                condition_str[0],
            )
        if order and order_by:
            query = "%s Order By %s %s;" % (query[:-1], order_by, order)

        try:
            self.cursor.execute(query)
            results = self.cursor.fetchall()
        except Exception as e:
            raise SQLError(str(e), query)

        if "bokeh_cds" in kwargs and kwargs["bokeh_cds"]:
            keys = [c.strip() for c in return_col_str.split(",")]
            results = {
                key: [results[r][i] for r in range(len(results))]
                for i, key in enumerate(keys)
            }
            for key in list(results):
                if "time" in key or "date" in key:
                    results[key] = [str(value) for value in results[key]]

        return results

    def query_generic(self, query_str):
        """A generic query function that executes the provided string

        Parameters
        ----------
        query_str : str
            SQL command

        Returns
        -------
        list
            Results of ``cursor.fetchall()``

        """
        self.cursor.execute(query_str)
        return self.cursor.fetchall()

    @property
    def now(self):
        """Get a datetime object for now

        Returns
        -------
        datetime
            The current time reported by the database

        """
        return self.query_generic("SELECT %s" % self.sql_cmd_now)[0][0]

    @property
    def sql_cmd_now(self):
        """Get the SQL command for now, based on database type

        Returns
        -------
        str
            SQL command for now
        """
        if self.db_type == "sqlite":
            sql_cmd = "strftime('%Y-%m-%d %H:%M:%f', 'now', 'localtime')"
        else:
            sql_cmd = "NOW()"
        return sql_cmd

    def update(self, table_name, column, value, condition_str):
        """Change the data in the database

        Parameters
        ----------
        table_name : str
            DVHs', 'Plans', 'Rxs', 'Beams', or 'DICOM_Files'
        column : str
            SQL column to be updated
        value : str, float, int
            value to be set
        condition_str : str
            a condition in SQL syntax

        """

        try:
            float(value)
            value_is_numeric = True
        except ValueError:
            value_is_numeric = False

        if "::date" in str(value):
            amend_type = ["", "::date"][
                self.db_type == "pgsql"
            ]  # sqlite3 does not support ::date
            value = "'%s'%s" % (
                value.strip("::date"),
                amend_type,
            )  # augment value for postgresql date formatting
        elif value_is_numeric:
            value = str(value)
        elif "null" == str(value.lower()):
            value = "NULL"
        else:
            value = "'%s'" % str(value)  # need quotes to input a string

        update = "Update %s SET %s = %s WHERE %s" % (
            table_name,
            column,
            value,
            condition_str,
        )

        try:
            self.cursor.execute(update)
            self.cnx.commit()
        except Exception as e:
            push_to_log(e, msg="Database update failure!")

    def is_study_instance_uid_in_table(self, table_name, study_instance_uid):
        """Check if a study instance uid exists in the provided table

        Parameters
        ----------
        table_name : str
            SQL table name
        study_instance_uid : str
            study instance uid

        Returns
        -------
        bool
            True if ``study_instance_uid`` exists in the study_instance_uid
            column of ``table_name``
        """
        # As of DVH v0.7.5, study_instance_uid may end with _N where N is the
        # nth plan of a file set
        query = (
            "SELECT DISTINCT study_instance_uid FROM %s WHERE "
            "study_instance_uid LIKE '%s%%';"
            % (table_name, study_instance_uid)
        )
        self.cursor.execute(query)
        results = self.cursor.fetchall()
        return bool(results)

    def is_mrn_in_table(self, table_name, mrn):
        """

        Parameters
        ----------
        table_name : str
            SQL table name
        mrn : str
            medical record number

        Returns
        -------
        bool
            True if ``mrn`` exists in the mrn column of ``table_name``
        """
        return self.is_value_in_table(table_name, mrn, "mrn")

    def is_value_in_table(self, table_name, value, column):
        """Check if a str value exists in a SQL table

        Parameters
        ----------
        table_name : sr
            SQL table name
        value : str
            value of interest (str only)
        column :
            SQL table column

        Returns
        -------
        bool
            True if ``value`` exists in ``table_name.column``

        """
        query = "Select Distinct %s from %s where %s = '%s';" % (
            column,
            table_name,
            column,
            value,
        )
        self.cursor.execute(query)
        results = self.cursor.fetchall()
        return bool(results)

    def insert_row(self, table, row):
        """Generic function to import data to the database

        Parameters
        ----------
        table : str
            SQL table name
        row : dict
            data returned from DICOM_Parser.get_<table>_row()

        """
        columns = list(row)
        allowed_columns = self.get_column_names(table)
        used_columns = []

        values = []
        for column in columns:
            if column in allowed_columns:
                used_columns.append(column)
                if (
                    row[column] is None
                    or row[column][0] is None
                    or row[column][0] == ""
                ):
                    if column == "import_time_stamp":
                        values.append(self.sql_cmd_now)
                    else:
                        values.append("NULL")
                else:
                    value = row[column][0]
                    value_type = row[column][1]

                    if "varchar" in value_type:
                        max_length = int(
                            value_type.replace("varchar(", "").replace(")", "")
                        )
                        values.append(
                            "'%s'" % truncate_string(value, max_length)
                        )

                    elif value_type in {"time_stamp", "date"}:
                        date = date_parser(value)
                        value = date.date()
                        if value_type == "time_stamp":
                            value = "%s %s" % (value, date.time())
                        if (
                            self.db_type == "pgsql"
                        ):  # sqlite3 does not support ::date
                            values.append("'%s'::date" % value)
                        else:
                            values.append("'%s'" % value)

                    else:
                        values.append("'%s'" % value)
            else:
                msg = (
                    "Failed to update SQL column %s in table %s, "
                    "it does not exist" % (column, table)
                )
                push_to_log(msg=msg)

        cmd = "INSERT INTO %s (%s) VALUES (%s);\n" % (
            table,
            ",".join(used_columns),
            ",".join(values),
        )
        self.execute_str(cmd)

    def insert_data_set(self, data_set):
        """Insert an entire data set for a plan

        Parameters
        ----------
        data_set : dict
            a dictionary of data with table names for keys, and a list of row
            data for values

        """
        for key in list(data_set):
            for row in data_set[key]:
                self.insert_row(key, row)

    def get_dicom_file_paths(self, mrn=None, uid=None):
        """Lookup the dicom file paths of imported data

        Parameters
        ----------
        mrn : str, optional
            medical record number
        uid : str, optional
            study instance uid

        Returns
        -------
        list
            Query return from DICOM_Files table

        """
        condition = None
        if uid:
            condition = "study_instance_uid = '%s'" % uid
        elif mrn:
            condition = "mrn = '%s'" % mrn

        if condition is not None:
            columns = (
                "mrn, study_instance_uid, folder_path, plan_file, "
                "structure_file, dose_file"
            )
            return self.query(
                "DICOM_Files", columns, condition, bokeh_cds=True
            )

    def delete_rows(self, condition_str, ignore_tables=None):
        """Delete all rows from all tables not in ignore_table for a given
        condition. Useful when deleting a plan/patient

        Parameters
        ----------
        condition_str : str: str
            a condition in SQL syntax
        ignore_tables : list, optional
            tables to be excluded from row deletions

        """

        tables = set(self.tables)
        if ignore_tables:
            tables = tables - set(ignore_tables)

        for table in tables:
            self.cursor.execute(
                "DELETE FROM %s WHERE %s;" % (table, condition_str)
            )
            self.cnx.commit()

    def change_mrn(self, old, new):
        """Edit all mrns in database

        Parameters
        ----------
        old : str
            current mrn
        new : str
            new mrn

        """
        condition = "mrn = '%s'" % old
        for table in self.tables:
            self.update(table, "mrn", new, condition)

    def change_uid(self, old, new):
        """Edit study instance uids in database

        Parameters
        ----------
        old : str
            current study instance uid
        new : str
            new study instance uid

        """
        condition = "study_instance_uid = '%s'" % old
        for table in self.tables:
            self.update(table, "study_instance_uid", new, condition)

    def delete_dvh(self, roi_name, study_instance_uid):
        """Delete a specified DVHs table row

        Parameters
        ----------
        roi_name : str
            the roi name for the row to be deleted
        study_instance_uid : str
            the associated study instance uid

        """
        self.cursor.execute(
            "DELETE FROM DVHs WHERE roi_name = '%s' and "
            "study_instance_uid = '%s';" % (roi_name, study_instance_uid)
        )
        self.cnx.commit()

    def ignore_dvh(self, variation, study_instance_uid, unignore=False):
        """Change an uncategorized roi name to ignored so that it won't show
        up in the list of uncategorized rois, so that
        the user doesn't have to evaluate its need everytime they cleanup the
        misc rois imported

        Parameters
        ----------
        variation : str
            roi name
        study_instance_uid : str
            the associated study instance uid
        unignore : bool, optional
            if set to True, sets the variation to 'uncategorized'

        """
        physician_roi = ["ignored", "uncategorized"][unignore]
        self.update(
            "dvhs",
            "physician_roi",
            physician_roi,
            "roi_name = '%s' and study_instance_uid = '%s'"
            % (variation, study_instance_uid),
        )

    def drop_tables(self):
        """Delete all tables in the database if they exist"""
        for table in self.tables:
            self.cursor.execute("DROP TABLE IF EXISTS %s;" % table)
            self.cnx.commit()

    def drop_table(self, table):
        """Delete a table in the database if it exists

        Parameters
        ----------
        table : str
            SQL table

        """
        self.cursor.execute("DROP TABLE IF EXISTS %s;" % table)
        self.cnx.commit()

    def initialize_database(self):
        """Ensure that all of the latest SQL columns exist in the database"""
        create_tables_file = [CREATE_PGSQL_TABLES, CREATE_SQLITE_TABLES][
            self.db_type == "sqlite"
        ]
        self.execute_file(create_tables_file)

    def reinitialize_database(self):
        """Delete all data and create all tables with latest columns"""
        self.drop_tables()
        self.vacuum()
        self.initialize_database()

    def vacuum(self):
        """Call to reclaim space in the database"""
        if self.db_type == "sqlite":
            self.cnx.isolation_level = None
            self.cnx.execute("VACUUM")
            self.cnx.isolation_level = ""
        else:
            # TODO: PGSQL VACUUM needs testing
            # self.cnx.execute('VACUUM')
            pass

    def does_db_exist(self):
        """Check if database exists

        Returns
        -------
        bool
            True if the database exists

        """
        if self.db_name:
            line = (
                "SELECT datname FROM pg_catalog.pg_database WHERE "
                "lower(datname) = lower('%s');" % self.db_name
            )
            self.cursor.execute(line)

            return bool(len(self.cursor.fetchone()))
        else:
            return True

    def is_sql_table_empty(self, table):
        """Check if specified SQL table is empty

        Parameters
        ----------
        table : str
            SQL table

        Returns
        -------
        bool
            True if ``table`` is empty

        """
        line = "SELECT COUNT(*) FROM %s;" % table
        self.cursor.execute(line)
        count = self.cursor.fetchone()[0]
        return not (bool(count))

    def get_unique_values(self, table, column, *condition, **kwargs):
        """Uses SELECT DISTINCT to get distinct values in database

        Parameters
        ----------
        table : str
            SQL table
        column : str
            SQL column
        condition : str, optional
            Condition in SQL syntax
        kwargs :
            option to ignore null values in return


        Returns
        -------
        list
            Unique values in table.column

        """
        if condition and condition[0]:
            query = "select distinct %s from %s where %s;" % (
                column,
                table,
                str(condition[0]),
            )
        else:
            query = "select distinct %s from %s;" % (column, table)
        self.cursor.execute(query)
        cursor_return = self.cursor.fetchall()
        if "ignore_null" in kwargs and kwargs["ignore_null"]:
            unique_values = [str(uv[0]) for uv in cursor_return if str(uv[0])]
        else:
            unique_values = [str(uv[0]) for uv in cursor_return]

        unique_values.sort()
        return unique_values

    def get_column_names(self, table_name):
        """Get all of the column names for a specified table

        Parameters
        ----------
        table_name : str
            SQL table

        Returns
        -------
        list
            All columns names in ``table_name``

        """
        if self.db_type == "sqlite":
            query = "PRAGMA table_info(%s);" % table_name.lower()
            index = 1
        else:
            query = (
                "select column_name from information_schema.columns where "
                "table_name = '%s';" % table_name.lower()
            )
            index = 0
        self.cursor.execute(query)
        cursor_return = self.cursor.fetchall()
        columns = [str(c[index]) for c in cursor_return]
        columns.sort()
        return columns

    def is_sqlite_column_datetime(self, table_name, column):
        """Check if a sqlite column is a datetime data type

        Parameters
        ----------
        table_name : str
            SQL table
        column : str
            SQL column

        Returns
        -------
        bool
            True if the ``table_name.column`` store datetime data

        """
        if self.db_type == "sqlite":
            query = "PRAGMA table_info(%s);" % table_name.lower()
            self.cursor.execute(query)
            cursor_return = self.cursor.fetchall()
            columns = {str(c[1]): str(c[2]) for c in cursor_return}
            if column in list(columns):
                column_type = columns[column]
                return (
                    "time" in column_type.lower()
                    or "date" in column_type.lower()
                )
        return False

    def get_min_value(self, table, column, condition=None):
        """Get the minimum value in the database for a given table and column

        Parameters
        ----------
        table : str
            SQL table
        column : str
            SQL column
        condition : str, optional
            Condition in SQL syntax

        Returns
        -------
        any
            The minimum value for table.column

        """
        return self.get_sql_function_value(
            "MIN", table, column, condition=condition
        )

    def get_max_value(self, table, column, condition=None):
        """Get the maximum value in the database for a given table and column

        Parameters
        ----------
        table : str
            SQL table
        column : str
            SQL column
        condition : str, optional
            Condition in SQL syntax

        Returns
        -------
        any
            The maximum value for table.column

        """
        return self.get_sql_function_value(
            "MAX", table, column, condition=condition
        )

    def get_sql_function_value(
        self, func, table, column, condition=None, first_value_only=True
    ):
        """Helper function used by get_min_values and get_max_values

        Parameters
        ----------
        func :
            SQL compatible function
        table : str
            SQL table
        column : str
            SQL column
        condition : str, optional
            Condition in SQL syntax (Default value = None)
        first_value_only : bool, optional
            if true, only return the first value, otherwise all values returned

        Returns
        -------
        list, any
            Results of ``cursor.fetchone()`` or ``cursor.fetchone()[0]`` if
            ``first_value_only`` is True

        """
        if condition:
            query = "SELECT %s(%s) FROM %s WHERE %s;" % (
                func,
                column,
                table,
                condition,
            )
        else:
            query = "SELECT %s(%s) FROM %s;" % (func, column, table)
        self.cursor.execute(query)
        cursor_return = self.cursor.fetchone()
        if first_value_only:
            return cursor_return[0]
        return cursor_return

    def get_roi_count_from_query(self, uid=None, dvh_condition=None):
        """Counts the DVH rows that match the provided conditions

        Parameters
        ----------
        uid : str, optional
            study instance uid
        dvh_condition : str, optional
            condition in SQL syntax for the DVHs table

        Returns
        -------
        int
            Number of rows in the DVHs table matching the provided parameters

        """
        if uid:
            condition = "study_instance_uid in ('%s')" % "', '".join(
                dvh_condition
            )
            if dvh_condition:
                condition = " and " + condition
        else:
            condition = ""

        if dvh_condition:
            condition = dvh_condition + condition

        return len(self.query("DVHs", "mrn", condition))

    def is_uid_imported(self, uid):
        """Check all tables to see if study instance uid is used

        Parameters
        ----------
        uid : str
            study instance uid

        Returns
        -------
        bool
            True if ``uid`` exists in any table

        """
        for table in self.tables:
            if self.is_study_instance_uid_in_table(table, uid):
                return True
        return False

    def is_mrn_imported(self, mrn):
        """Check all tables to see if MRN is used

        Parameters
        ----------
        mrn : str
            medical record number

        Returns
        -------
        bool
            True if ``mrn`` exists in any table

        """
        for table in self.tables:
            if self.is_mrn_in_table(table, mrn):
                return True
        return False

    def is_roi_imported(self, roi_name, study_instance_uid):
        """Check if a study is already using a specified roi name

        Parameters
        ----------
        roi_name : str
            roi name to check
        study_instance_uid : str
            restrict search to this study_instance_uid

        Returns
        -------
        bool
            True if ``roi_name`` is used in the DVHs table for the given
            ``study_instance_uid``

        """
        condition = "roi_name = '%s' and study_instance_uid = '%s'" % (
            roi_name,
            study_instance_uid,
        )
        roi_names = self.get_unique_values("DVHs", "roi_name", condition)
        return bool(roi_names)

    def get_row_count(self, table, condition=None):
        """

        Parameters
        ----------
        table : str
            SQL table
        condition : str
            SQL condition

        Returns
        -------
        int
            Number of rows in ``table`` meeting ``condition``

        """
        ans = self.query(table, "COUNT(mrn)", condition)
        if ans:
            return ans[0][0]
        return 0

    def export_to_sqlite(self, file_path, callback=None, force=False):
        """Create a new SQLite database and import this database's data

        Parameters
        ----------
        file_path : str
            Path where the new SQLite database will be created
        callback : callable, optional
            optional function to be called on each row insertion. Should accept
            table (str), current row (int), total_row_count (int) as parameters
        force : bool, optional
            ignore duplicate StudyInstanceUIDs if False

        """
        config = {"host": file_path}
        new_cnx = DVH_SQL(config, db_type="sqlite")
        new_cnx.initialize_database()
        self.import_db(self, new_cnx, callback=callback, force=force)

    @staticmethod
    def import_db(cnx_src, cnx_dst, callback=None, force=False):
        """

        Parameters
        ----------
        cnx_src : DVH_SQL
            the source DVHA DB connection
        cnx_dst : DVH_SQL
            the destination DVHA DB connection
        callback : callable, optional
            optional function to be called on each row insertion. Should accept
            table (str), current row (int), total_row_count (int) as parameters
        force : bool, optional
            ignore duplicate StudyInstanceUIDs if False

        """
        for table in cnx_src.tables:
            columns = cnx_src.get_column_names(table)
            study_uids_1 = cnx_src.get_unique_values(
                table, "study_instance_uid"
            )

            condition = None
            if not force:
                study_uids_2 = cnx_dst.get_unique_values(
                    table, "study_instance_uid"
                )
                study_uids_1 = list(set(study_uids_1) - set(study_uids_2))
                condition = "study_instance_uid IN ('%s')" % "','".join(
                    study_uids_1
                )

            total_row_count = cnx_src.get_row_count(table, condition)

            counter = 0
            for uid in study_uids_1:
                study_data = cnx_src.query(
                    table, ",".join(columns), "study_instance_uid = '%s'" % uid
                )
                for row in study_data:
                    if callback is not None:
                        CallAfter(callback, table, counter, total_row_count)
                        counter += 1
                    row_str = "'" + "','".join([str(v) for v in row]) + "'"
                    row_str = row_str.replace("'None'", "NULL")
                    cmd = "INSERT INTO %s (%s) VALUES (%s);\n" % (
                        table,
                        ",".join(columns),
                        row_str,
                    )
                    cnx_dst.execute_str(cmd)

    def save_to_json(self, file_path, callback=None):
        """Export SQL database to a JSON file

        Parameters
        ----------
        file_path : str
            file_path to new JSON file
        callback : callable, optional
            optional function to be called on each table insertion. Should
            accept table_name (str), table (int), table_count (int) as
            parameters

        """
        json_data = {
            "columns": {"categorical": categorical, "numerical": numerical}
        }
        for i, table in enumerate(self.tables):
            if callback is not None:
                CallAfter(callback, table, i, len(self.tables))
            columns = self.get_column_names(table)
            json_data[table] = self.query(
                table, ",".join(columns), bokeh_cds=True
            )

        with open(file_path, "w") as fp:
            json.dump(json_data, fp)

    def get_ptv_counts(self):
        """Get number of PTVs for each study instance uid

        Returns
        -------
        dict
            PTV counts stored by ``study_instance_uid``

        """
        results = self.query(
            "DVHs", "study_instance_uid", "roi_type like 'PTV%'"
        )
        uids = [uid[0] for uid in results]
        return {uid: uids.count(uid) for uid in list(set(uids))}


def truncate_string(input_string, character_limit):
    """Used to truncate a string to ensure it may be imported into database

    Parameters
    ----------
    input_string : str
        string to be truncated
    character_limit : int
        the maximum number of allowed characters

    Returns
    -------
    str
        truncated string

    """
    if len(input_string) > character_limit:
        return input_string[0 : (character_limit - 1)]
    return input_string


def echo_sql_db(config=None, db_type="pgsql", group=1):
    """Echo the database using stored or provided credentials

    Parameters
    ----------
    config : dict, optional
        database login credentials
    db_type : str, optional
        either 'pgsql' or 'sqlite'
    group : int, optional
        either group 1 or 2

    Returns
    -------
    bool
        True if echo is successful

    """
    try:
        if config:
            if db_type == "pgsql" and (
                "dbname" not in list(config) or "port" not in list(config)
            ):
                return False
            cnx = DVH_SQL(config, db_type=db_type, group=group)
        else:
            cnx = DVH_SQL(group=group)
        cnx.close()
        return True
    except Exception as e:
        if type(e) not in [
            psycopg2.OperationalError,
            sqlite3.OperationalError,
        ]:
            push_to_log(e, msg="Unknown Error during SQL Echo")
        return False


def write_test(
    config=None, db_type="pgsql", group=1, table=None, column=None, value=None
):
    """Write test data to database, verify with a query

    Parameters
    ----------
    config : dict, optional
        database login credentials
    db_type : str, optional
        either 'pgsql' or 'sqlite'
    group : int, optional
        either group 1 or 2
    table : str, optional
        SQL table
    column : str, optional
        SQL column
    value : str, optional
        test value

    Returns
    -------
    dict
        Write and Read test statuses

    """

    try:
        if config:
            if db_type == "pgsql" and (
                "dbname" not in list(config) or "port" not in list(config)
            ):
                return None
            cnx = DVH_SQL(config, db_type=db_type, group=group)
        else:
            cnx = DVH_SQL(group=group)
    except Exception as e:
        push_to_log(
            e, msg="Write Test: Connection to SQL could not be established"
        )
        return {"write": False, "delete": False}

    try:
        cnx.initialize_database()
    except Exception as e:
        push_to_log(e, msg="Write Test: DVH_SQL.initialize_database failed")

    if table is None:
        table = cnx.tables[-1]

    if column is None:
        column = "mrn"

    if value is None:
        # Find a test value that does not exist in the database
        value_init = "SqlTest_"
        i = 0
        value = value_init + str(i)
        current_values = cnx.get_unique_values(table, column)
        while value in current_values:
            value = value_init + str(i)
            i += 1

    condition_str = "%s = '%s'" % (column, value)
    insert_cmd = "INSERT INTO %s (%s) VALUES ('%s');" % (table, column, value)
    delete_cmd = "DELETE FROM %s WHERE %s;" % (table, condition_str)

    try:
        cnx.execute_str(insert_cmd)
    except Exception as e:
        push_to_log(e, msg="Write Test: SQL test insert command failed")

    try:
        test_return = cnx.query(table, column, condition_str)
        write_test_success = len(test_return) > 0
    except Exception as e:
        write_test_success = False
        push_to_log(e, msg="Write Test: SQL query of test insert failed")

    if not write_test_success:
        delete_test_success = None
    else:
        try:
            cnx.execute_str(delete_cmd)
            test_return = cnx.query(table, column, condition_str)
            delete_test_success = len(test_return) == 0
        except Exception as e:
            delete_test_success = False
            push_to_log(
                e, msg="Write Test: SQL delete command of test insert failed"
            )

    try:
        cnx.close()
    except Exception as e:
        push_to_log(e, msg="Write Test: Failed to close SQL connection")

    return {"write": write_test_success, "delete": delete_test_success}


def initialize_db():
    """Initialize the database"""
    with DVH_SQL() as cnx:
        cnx.initialize_database()


def is_file_sqlite_db(sqlite_db_file):
    """Check if file is a sqlite database

    Parameters
    ----------
    sqlite_db_file : str
        path to file to be checked


    Returns
    -------
    bool
        True if ``sqlite_db_file`` is a sqlite database

    """
    if isfile(sqlite_db_file):
        try:
            cnx = sqlite3.connect(sqlite_db_file)
            cnx.close()
            return True
        except sqlite3.OperationalError:
            pass

    return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# tools.utilties.py
"""
General utilities for DVHA

"""
# Copyright (c) 2016-2019 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVH-Analytics

import wx
from datetime import datetime
from dateutil.parser import parse as parse_date
import linecache
import numpy as np
from os import walk, listdir, unlink, mkdir, rmdir, chdir, sep, environ
from os.path import (
    join,
    isfile,
    isdir,
    splitext,
    basename,
    dirname,
    realpath,
    pathsep,
)
import pickle
import pydicom
from pydicom.uid import ImplicitVRLittleEndian
import shutil
from subprocess import check_output
import sys
import tracemalloc
from dvha.db.sql_connector import DVH_SQL
from dvha.paths import (
    SQL_CNF_PATH,
    WIN_APP_ICON,
    PIP_LIST_PATH,
    DIRECTORIES,
    APP_DIR,
    BACKUP_DIR,
    DATA_DIR,
)
from dvha.tools.errors import push_to_log


IGNORED_FILES = [".ds_store"]

if environ.get("READTHEDOCS") == "True" or "sphinx" in sys.prefix:
    MSG_DLG_FLAGS = None
else:
    MSG_DLG_FLAGS = wx.ICON_WARNING | wx.YES | wx.NO | wx.NO_DEFAULT


def is_windows():
    """ """
    return wx.Platform == "__WXMSW__"


def set_msw_background_color(window_obj, color="lightgrey"):
    """

    Parameters
    ----------
    window_obj :

    color :
         (Default value = 'lightgrey')

    Returns
    -------

    """
    if is_windows():
        window_obj.SetBackgroundColour(color)


def is_linux():
    """ """
    return wx.Platform == "__WXGTK__"


def is_mac():
    """ """
    return wx.Platform == "__WXMAC__"


def initialize_directories():
    """Based on paths.py, create required directories if they do not exist"""
    for directory in DIRECTORIES.values():
        if not isdir(directory):
            mkdir(directory)


def write_sql_connection_settings(config):
    """Create a file storing the SQL login credentials

    Parameters
    ----------
    config : dict
        contains values for 'host', 'dbname', 'port' and optionally 'user' and 'password'

    Returns
    -------

    """
    # TODO: Make this more secure

    text = ["%s %s" % (key, value) for key, value in config.items() if value]
    text = "\n".join(text)

    with open(SQL_CNF_PATH, "w") as text_file:
        text_file.write(text)


def scale_bitmap(bitmap, width, height):
    """Used to scale tool bar images for MSW and GTK, MAC automatically scales

    Parameters
    ----------
    bitmap :
        bitmap to be scaled
        type bitmap: Bitmap
    width : int
        width of output bitmap
    height : int
        height of output bitmap

    Returns
    -------
    Bitmap
        scaled bitmap

    """
    image = wx.Bitmap.ConvertToImage(bitmap)
    image = image.Scale(width, height, wx.IMAGE_QUALITY_HIGH)
    return wx.Bitmap(image)


def get_tree_ctrl_image(
    file_path, file_type=wx.BITMAP_TYPE_PNG, width=16, height=16
):
    """Create an image top be used in the TreeCtrl from the provided file_path

    Parameters
    ----------
    file_path : str
        absolute file_path of image
    file_type :
        specify the image format (PNG by default)
    width : int
        width of output bitmap (16 default)
    height : int
        height of output bitmap (16 default)

    Returns
    -------
    Image
        scaled image for TreeCtrl

    """
    return (
        wx.Image(file_path, file_type).Scale(width, height).ConvertToBitmap()
    )


def get_file_paths(
    start_path, search_subfolders=False, extension=None, return_dict=False
):
    """Get a list of absolute file paths for a given directory

    Parameters
    ----------
    start_path : str
        initial directory
    search_subfolders : bool
        optionally search all sub folders (Default value = False)
    extension : str
        optionally include only files with specified extension (Default value = None)
    return_dict :
         (Default value = False)

    Returns
    -------
    list or dict
        absolute file paths

    """
    if isdir(start_path):
        file_paths = []
        file_paths_dict = {}
        if search_subfolders:
            for root, dirs, files in walk(start_path, topdown=False):
                for name in files:
                    if (
                        extension is None
                        or splitext(name)[1].lower() == extension.lower()
                    ):
                        if name.lower() not in IGNORED_FILES:
                            file_paths.append(join(root, name))
                            if return_dict:
                                if root not in file_paths_dict.keys():
                                    file_paths_dict[root] = []
                                file_paths_dict[root].append(name)
            if return_dict:
                return file_paths_dict
            return file_paths

        for f in listdir(start_path):
            if isfile(join(start_path, f)):
                if (
                    extension is None
                    or splitext(f)[1].lower() == extension.lower()
                ):
                    if f.lower() not in IGNORED_FILES:
                        file_paths.append(join(start_path, f))
        return file_paths
    return []


def get_study_instance_uids(**kwargs):
    """Get lists of study instance uids in the SQL database that meet provided conditions
    The values return in the 'common' key are used for the DVH class in models.dvh.py

    Parameters
    ----------
    kwargs :
        keys are SQL table names and the values are conditions in SQL syntax
    **kwargs :


    Returns
    -------
    dict
        study instance uids for each table, uids found in all tables, and a list of unique uids

    """
    with DVH_SQL() as cnx:
        uids = {
            table: cnx.get_unique_values(
                table, "study_instance_uid", condition
            )
            for table, condition in kwargs.items()
        }

    complete_list = flatten_list_of_lists(
        list(uids.values()), remove_duplicates=True
    )

    uids["common"] = [
        uid for uid in complete_list if is_uid_in_all_keys(uid, uids)
    ]
    uids["unique"] = complete_list

    return uids


def is_uid_in_all_keys(uid, uids):
    """Check if uid is found in each of the uid lists for each SQL table

    Parameters
    ----------
    uid : str
        study instance uid
    uids : dict
        lists of study instance uids organized by SQL table

    Returns
    -------
    bool
        True only if uid is found in each of the tables

    """

    table_answer = {}
    # Initialize a False value for each key
    for table in list(uids):
        table_answer[table] = False
    # search for uid in each keyword fof uid_kwlist
    for table, value in uids.items():
        if uid in value:
            table_answer[table] = True

    final_answer = True
    # Product of all answer[key] values (except 'unique')
    for table, value in table_answer.items():
        if table not in "unique":
            final_answer *= value
    return final_answer


def flatten_list_of_lists(some_list, remove_duplicates=False, sort=False):
    """Convert a list of lists into a list of all values

    Parameters
    ----------
    some_list : list: list
        a list such that each value is a list
    remove_duplicates : bool
        if True, return a unique list, otherwise keep duplicated values (Default value = False)
    sort : bool
        if True, sort the list (Default value = False)

    Returns
    -------
    type
        a new object containing all values in teh provided

    """
    data = [item for sublist in some_list for item in sublist]

    if remove_duplicates:
        if sort:
            return list(set(data))
        else:
            ans = []
            for value in data:
                if value not in ans:
                    ans.append(value)
            return ans
    elif sort:
        return sorted(data)

    return data


def collapse_into_single_dates(x, y):
    """Function used for a time plot to convert multiple values into one value, while retaining enough information
    to perform a moving average over time

    Parameters
    ----------
    x :
        a list of dates in ascending order
    y :
        a list of values and can use the '+' operator as a function of date

    Returns
    -------
    dict
        a unique list of dates, sum of y for that date, and number of original points for that date

    """

    # average daily data and keep track of points per day
    x_collapsed = [x[0]]
    y_collapsed = [y[0]]
    w_collapsed = [1]
    for n in range(1, len(x)):
        if x[n] == x_collapsed[-1]:
            y_collapsed[-1] = y_collapsed[-1] + y[n]
            w_collapsed[-1] += 1
        else:
            x_collapsed.append(x[n])
            y_collapsed.append(y[n])
            w_collapsed.append(1)

    return {"x": x_collapsed, "y": y_collapsed, "w": w_collapsed}


def moving_avg(xyw, avg_len):
    """Calculate a moving average for a given averaging length

    Parameters
    ----------
    xyw : dict
        output from collapse_into_single_dates
    avg_len : int
        average of these number of points, i.e., look-back window

    Returns
    -------
    tuple
        list of x values, list of y values

    """
    cumsum, moving_aves, x_final = [0], [], []

    for i, y in enumerate(xyw["y"], 1):
        cumsum.append(cumsum[i - 1] + y / xyw["w"][i - 1])
        if i >= avg_len:
            moving_ave = (cumsum[i] - cumsum[i - avg_len]) / avg_len
            moving_aves.append(moving_ave)
    x_final = [xyw["x"][i] for i in range(avg_len - 1, len(xyw["x"]))]

    return x_final, moving_aves


def convert_value_to_str(value, rounding_digits=2):
    """

    Parameters
    ----------
    value :

    rounding_digits :
         (Default value = 2)

    Returns
    -------

    """
    try:
        formatter = "%%0.%df" % rounding_digits
        return formatter % value
    except TypeError:
        return value


def get_selected_listctrl_items(list_control):
    """Get the indices of the currently selected items of a wx.ListCtrl object

    Parameters
    ----------
    list_control : ListCtrl
        any wx.ListCtrl object

    Returns
    -------
    list
        indices of selected items

    """
    selection = []

    index_current = -1
    while True:
        index_next = list_control.GetNextItem(
            index_current, wx.LIST_NEXT_ALL, wx.LIST_STATE_SELECTED
        )
        if index_next == -1:
            return selection

        selection.append(index_next)
        index_current = index_next


def print_run_time(start_time, end_time, calc_title):
    """

    Parameters
    ----------
    start_time : datetime
        start time of process
    end_time : datetime
        end time of process
    calc_title : str
        prepend the status message with this value

    Returns
    -------

    """
    total_time = end_time - start_time
    seconds = total_time.seconds
    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)
    if h:
        print(
            "%s. This took %dhrs %02dmin %02dsec to complete"
            % (calc_title, h, m, s)
        )
    elif m:
        print("%s. This took %02dmin %02dsec to complete" % (calc_title, m, s))
    else:
        print("%s. This took %02dsec to complete" % (calc_title, s))


def datetime_to_date_string(datetime_obj):
    """

    Parameters
    ----------
    datetime_obj :


    Returns
    -------

    """
    if isinstance(datetime_obj, str):
        datetime_obj = parse_date(datetime_obj)
    return "%s/%s/%s" % (
        datetime_obj.month,
        datetime_obj.day,
        datetime_obj.year,
    )


def change_angle_origin(angles, max_positive_angle):
    """Angles in DICOM are all positive values, but there is typically no mechanical continuity in across 180 degrees

    Parameters
    ----------
    angles : list
        angles to be converted
    max_positive_angle :
        the maximum positive angle, angles greater than this will be shifted to negative angles

    Returns
    -------
    list
        list of the same angles, but none exceed the max

    """
    if len(angles) == 1:
        if angles[0] > max_positive_angle:
            return [angles[0] - 360]
        else:
            return angles
    new_angles = []
    for angle in angles:
        if angle > max_positive_angle:
            new_angles.append(angle - 360)
        elif angle == max_positive_angle:
            if angle == angles[0] and angles[1] > max_positive_angle:
                new_angles.append(angle - 360)
            elif angle == angles[-1] and angles[-2] > max_positive_angle:
                new_angles.append(angle - 360)
            else:
                new_angles.append(angle)
        else:
            new_angles.append(angle)
    return new_angles


def calc_stats(data):
    """Calculate a standard set of stats for DVHA

    Parameters
    ----------
    data : list
        a list or numpy 1D array of numbers

    Returns
    -------
    list
        max, 75%, median, mean, 25%, and min of data

    """
    data = [x for x in data if x != "None"]
    try:
        data_np = np.array(data)
        rtn_data = [
            np.max(data_np),
            np.percentile(data_np, 75),
            np.median(data_np),
            np.mean(data_np),
            np.percentile(data_np, 25),
            np.min(data_np),
        ]
    except Exception as e:
        rtn_data = [0, 0, 0, 0, 0, 0]
        msg = "tools.utilities.calc_stats: received non-numerical data"
        push_to_log(e, msg=msg)
    return rtn_data


def move_files_to_new_path(
    files, new_dir, copy_files=False, new_file_names=None, callback=None
):
    """Move all files provided to the new directory

    Parameters
    ----------
    files : list
        absolute file paths
    new_dir : str
        absolute directory path
    copy_files : bool
        Set to True to keep original files and copy to new_dir, False to remove original files (Default value = False)
    new_file_names : None or list of str
        optionally provide a list of new names (Default value = None)
    callback : callable
        optional function to call at the start of each iteration (Default value = None)

    Returns
    -------

    """
    for i, file_path in enumerate(files):
        if callback is not None:
            callback(i, len(files))
        if isfile(file_path):
            file_name = basename(file_path)
            old_dir = dirname(file_path)
            new_file_name = (
                file_name if new_file_names is None else new_file_names[i]
            )
            new = join(new_dir, new_file_name)
            if not isdir(new_dir):
                mkdir(new_dir)
            if old_dir != new_dir:
                if isfile(new):
                    if not copy_files:
                        delete_file(new)
                else:
                    [shutil.move, shutil.copy][copy_files](file_path, new)


def delete_directory_contents(dir_to_delete):
    """

    Parameters
    ----------
    dir_to_delete :


    Returns
    -------

    """
    # https://stackoverflow.com/questions/185936/how-to-delete-the-contents-of-a-folder-in-python
    for the_file in listdir(dir_to_delete):
        delete_file(join(dir_to_delete, the_file))


def delete_file(file_path):
    """

    Parameters
    ----------
    file_path :


    Returns
    -------

    """
    try:
        if isfile(file_path):
            unlink(file_path)
        elif isdir(file_path):
            shutil.rmtree(file_path)
    except Exception as e:
        push_to_log(e, msg="tools.utilities.delete_file: %s" % file_path)


def delete_imported_dicom_files(dicom_files):
    """delete imported dicom files

    Parameters
    ----------
    dicom_files : dict
        the return from DVH_SQL().get_dicom_file_paths

    Returns
    -------

    """
    for i, directory in enumerate(dicom_files["folder_path"]):
        # Delete associated plan, structure, and dose files
        for key in ["plan_file", "structure_file", "dose_file"]:
            delete_file(join(directory, dicom_files[key][i]))

        # Delete misc dicom files for given study instance uid
        remaining_files = listdir(directory)
        for f in remaining_files:
            try:
                uid = str(
                    pydicom.read_file(join(directory, f)).StudyInstanceUID
                )
                if uid == str(dicom_files["study_instance_uid"][i]):
                    delete_file(f)
            except Exception:
                pass

        # Directory is empty, delete it
        # Directories are by patient mrn, so it might contain files for a different study for the same patient
        if not listdir(directory):
            try:
                rmdir(directory)
            except Exception:
                pass


def move_imported_dicom_files(dicom_files, new_dir):
    """move imported dicom files

    Parameters
    ----------
    dicom_files : dict
        the return from DVH_SQL().get_dicom_file_paths
    new_dir :


    Returns
    -------

    """
    for i, directory in enumerate(dicom_files["folder_path"]):
        files = [
            join(directory, dicom_files[key][i])
            for key in ["plan_file", "structure_file", "dose_file"]
        ]
        new_patient_dir = join(new_dir, dicom_files["mrn"][i])
        move_files_to_new_path(files, new_patient_dir)

        # Move misc dicom files for given study instance uid
        remaining_files = listdir(directory)
        files = []
        for f in remaining_files:
            try:
                uid = str(
                    pydicom.read_file(join(directory, f)).StudyInstanceUID
                )
                if uid == str(dicom_files["study_instance_uid"][i]):
                    files.append(f)
            except Exception:
                pass
        move_files_to_new_path(files, new_patient_dir)

        # Directory is empty, delete it
        # Directories are by patient mrn, so it might contain files for a different study for the same patient
        if not listdir(directory):
            try:
                rmdir(directory)
            except Exception:
                pass


def remove_empty_sub_folders(start_path):
    """

    Parameters
    ----------
    start_path :


    Returns
    -------

    """
    for (path, dirs, files) in walk(start_path, topdown=False):
        if files:
            continue
        try:
            if path != start_path:
                rmdir(path)
        except OSError:
            pass


def move_all_files(new_dir, old_dir):
    """This function will move all files from the old to new directory, it will ignore all files in subdirectories

    Parameters
    ----------
    new_dir :
        absolute directory path
    old_dir :
        absolute directory path

    Returns
    -------

    """
    initial_path = dirname(realpath(__file__))

    chdir(old_dir)

    file_paths = [f for f in listdir(old_dir) if isfile(join(old_dir, f))]

    misc_path = join(new_dir, "misc")
    if not isdir(misc_path):
        mkdir(misc_path)

    for f in file_paths:
        file_name = basename(f)
        new = join(misc_path, file_name)
        shutil.move(f, new)

    chdir(initial_path)


def get_elapsed_time(start_time, end_time):
    """

    Parameters
    ----------
    start_time :

    end_time :


    Returns
    -------

    """
    total_time = end_time - start_time
    seconds = total_time.seconds
    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)
    if h:
        return "%d hrs %d min %d sec" % (h, m, s)
    if m:
        return "%d min %d sec" % (m, s)
    return "%d sec" % s


def is_date(date):
    """

    Parameters
    ----------
    date :


    Returns
    -------

    """
    if isinstance(date, datetime):
        return True

    if isinstance(date, str):
        try:
            parse_date(date)
            return True
        except Exception:
            return False

    return False


def rank_ptvs_by_D95(ptvs):
    """Determine the order of provided PTVs by their D_95% values

    Parameters
    ----------
    ptvs : dict
        dvh, volume, index of PTVs

    Returns
    -------
    type
        ptv numbers in order of D_95%

    """
    doses_to_rank = get_dose_to_volume(ptvs["dvh"], ptvs["volume"], 0.95)
    return sorted(range(len(ptvs["dvh"])), key=lambda k: doses_to_rank[k])


def get_dose_to_volume(dvhs, volumes, roi_fraction):
    """

    Parameters
    ----------
    dvhs :

    volumes :

    roi_fraction :


    Returns
    -------

    """
    # Not precise (i.e., no interpolation) but good enough for sorting PTVs
    doses = []
    for i, dvh in enumerate(dvhs):
        abs_volume = volumes[i] * roi_fraction
        dvh_np = np.array(dvh.split(","), dtype=np.float)
        try:
            dose = next(x[0] for x in enumerate(dvh_np) if x[1] < abs_volume)
        except StopIteration:
            dose = dvh_np[-1]
        doses.append(dose)

    return doses


def float_or_none(value):
    """

    Parameters
    ----------
    value :


    Returns
    -------

    """
    try:
        return float(value)
    except ValueError:
        return "None"


class MessageDialog:
    """This is the base class for Yes/No Dialog boxes
    Inherit this class, then over-write action_yes and action_no functions with appropriate behaviors

    Parameters
    ----------

    Returns
    -------

    """

    def __init__(
        self,
        parent,
        caption,
        message="Are you sure?",
        action_yes_func=None,
        action_no_func=None,
        flags=MSG_DLG_FLAGS,
    ):
        if is_windows():
            message = "\n".join([caption, message])
            caption = " "
        self.dlg = wx.MessageDialog(parent, message, caption, flags)
        self.parent = parent
        self.action_yes_func = action_yes_func
        self.action_no_func = action_no_func
        self.run()

    def run(self):
        """ """
        res = self.dlg.ShowModal()
        [self.action_no, self.action_yes][res == wx.ID_YES]()
        self.dlg.Destroy()

    def action_yes(self):
        """ """
        if self.action_yes_func is not None:
            self.action_yes_func()

    def action_no(self):
        """ """
        if self.action_no_func is not None:
            self.action_no_func()


def save_object_to_file(obj, abs_file_path):
    """Save a python object acceptable for pickle to the provided file path

    Parameters
    ----------
    obj :

    abs_file_path :


    Returns
    -------

    """
    with open(abs_file_path, "wb") as outfile:
        pickle.dump(obj, outfile)


def load_object_from_file(abs_file_path):
    """Load a pickled object from the provided absolute file path

    Parameters
    ----------
    abs_file_path :


    Returns
    -------

    """
    if isfile(abs_file_path):
        with open(abs_file_path, "rb") as infile:
            obj = pickle.load(infile)
        return obj


def sample_list(some_list, max_size, n):
    """Reduce a list by given factor iteratively until list size less than max_size

    Parameters
    ----------
    some_list : list: list
        any list you like!
    max_size : int
        the maximum number of items in the returned list
    n : int
        remove every nth element

    Returns
    -------
    list
        sampled list

    """
    while len(some_list) > max_size:
        some_list = remove_every_nth_element(some_list, n)
    return some_list


def remove_every_nth_element(some_list, n):
    """

    Parameters
    ----------
    some_list :

    n :


    Returns
    -------

    """
    return [value for i, value in enumerate(some_list) if i % n != 0]


def sample_roi(roi_coord, max_point_count=5000, iterative_reduction=0.1):
    """Iteratively sample a list of 3D points by the iterative_reduction until the size of the list is < max_point_count
    This is used to reduce the number of points used in the ptv distance calculations because:
    1) Shapely returns a much large number of points when calculating total PTVs
    2) Users could easily run into memory issues using scip.dist if all points are used (particularly on MSW)

    Parameters
    ----------
    roi_coord : list
        a list of 3D points representing an roi
    max_point_count : int_count: int
        the maximum number of points in the returned roi_coord (Default value = 5000)
    iterative_reduction : float
        iteratively remove this fraction of points until len < max_point_count (Default value = 0.1)

    Returns
    -------
    list
        sampled roi

    """
    return sample_list(
        roi_coord, max_point_count, int(1 / iterative_reduction)
    )


def get_sorted_indices(some_list):
    """

    Parameters
    ----------
    some_list :


    Returns
    -------

    """
    try:
        return [i[0] for i in sorted(enumerate(some_list), key=lambda x: x[1])]
    except TypeError:  # can't sort if a mix of str and float
        try:
            temp_data = [
                [value, -float("inf")][value == "None"] for value in some_list
            ]
            return [
                i[0] for i in sorted(enumerate(temp_data), key=lambda x: x[1])
            ]
        except TypeError:
            temp_data = [str(value) for value in some_list]
            return [
                i[0] for i in sorted(enumerate(temp_data), key=lambda x: x[1])
            ]


def get_window_size(width, height):
    """Function used to adapt frames/windows for the user's resolution

    Parameters
    ----------
    width :
        fractional width of the user's screen
    height :
        fractional height of the user's screen

    Returns
    -------
    tuple
        window size

    """
    user_width, user_height = wx.GetDisplaySize()
    if user_width / user_height < 1.5:  # catch 4:3 or non-widescreen
        user_height = user_width / 1.6
    return tuple([int(width * user_width), int(height * user_height)])


def apply_resolution_limits(size, options):
    """Apply min and max limits to a window size

    Parameters
    ----------
    size : tuple
        window size (width, height)
    options : Options
        dvha.options.Options class object

    Returns
    -------
    tuple
        width, height
    """

    min_size = options.MIN_RESOLUTION_MAIN
    max_size = options.MAX_INIT_RESOLUTION_MAIN
    return tuple([min(max(size[i], min_size[i]), max_size[i]) for i in [0, 1]])


def set_frame_icon(frame):
    """

    Parameters
    ----------
    frame :


    Returns
    -------

    """
    if not is_mac():
        frame.SetIcon(wx.Icon(WIN_APP_ICON))


def trace_memory_alloc_pretty_top(snapshot, key_type="lineno", limit=10):
    """From https://docs.python.org/3/library/tracemalloc.html

    Parameters
    ----------
    snapshot :

    key_type :
         (Default value = 'lineno')
    limit :
         (Default value = 10)

    Returns
    -------

    """
    print("-----------------------------------------------------------------")
    snapshot = snapshot.filter_traces(
        (
            tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),
            tracemalloc.Filter(False, "<unknown>"),
        )
    )
    top_stats = snapshot.statistics(key_type)

    print("Top %s lines" % limit)
    for index, stat in enumerate(top_stats[:limit], 1):
        frame = stat.traceback[0]
        # replace "/path/to/module/file.py" with "module/file.py"
        filename = sep.join(frame.filename.split(sep)[-2:])
        print(
            "#%s: %s:%s: %.1f KiB"
            % (index, filename, frame.lineno, stat.size / 1024)
        )
        line = linecache.getline(frame.filename, frame.lineno).strip()
        if line:
            print("    %s" % line)

    other = top_stats[limit:]
    if other:
        size = sum(stat.size for stat in other)
        print("%s other: %.1f KiB" % (len(other), size / 1024))
    total = sum(stat.size for stat in top_stats)
    print("Total allocated size: %.1f KiB" % (total / 1024))


def trace_memory_alloc_simple_stats(snapshot, key_type="lineno"):
    """From https://docs.python.org/3/library/tracemalloc.html

    Parameters
    ----------
    snapshot :

    key_type :
         (Default value = 'lineno')

    Returns
    -------

    """
    print("-----------------------------------------------------------------")
    print("[ Top 10 ]")
    top_stats = snapshot.statistics(key_type)
    for stat in top_stats[:10]:
        print(stat)


class PopupMenu:
    """ """

    def __init__(self, parent):
        self.parent = parent
        self.menus = []

    def add_menu_item(self, label, action):
        """

        Parameters
        ----------
        label :

        action :


        Returns
        -------

        """
        self.menus.append({"id": wx.NewId(), "label": label, "action": action})

    def run(self):
        """ """
        popup_menu = wx.Menu()
        for menu in self.menus:
            popup_menu.Append(menu["id"], menu["label"])
            self.parent.Bind(wx.EVT_MENU, menu["action"], id=menu["id"])

        self.parent.PopupMenu(popup_menu)
        popup_menu.Destroy()


def validate_transfer_syntax_uid(data_set):
    """

    Parameters
    ----------
    data_set :


    Returns
    -------

    """
    meta = pydicom.Dataset()
    meta.ImplementationClassUID = pydicom.uid.generate_uid()
    meta.TransferSyntaxUID = ImplicitVRLittleEndian
    new_data_set = pydicom.FileDataset(
        filename_or_obj=None,
        dataset=data_set,
        is_little_endian=True,
        file_meta=meta,
    )
    new_data_set.is_little_endian = True
    new_data_set.is_implicit_VR = True

    return new_data_set


def get_installed_python_libraries():
    """Use pip command line function 'list' to extract the currently installed libraries"""

    if isfile(PIP_LIST_PATH):  # Load a frozen list of packages if stored
        return load_object_from_file(PIP_LIST_PATH)
    try:  # If running from PyInstaller, this may fail, pickle a file prior to freezing with save_pip_list
        output = str(check_output(["pip", "list", "--local"]), "utf-8").split(
            "\n"
        )
    except Exception:
        return load_object_from_file(PIP_LIST_PATH)

    python_version = ".".join(str(i) for i in sys.version_info[:3])
    libraries = {"Library": ["python"], "Version": [python_version]}
    for row in output[
        2:
    ]:  # ignore first two rows which are column headers and a separator
        data = [v for v in row.strip().split(" ") if v]
        if data:
            libraries["Library"].append(data[0])
            libraries["Version"].append(data[1])
    return libraries


def save_pip_list():
    """ """
    save_object_to_file(get_installed_python_libraries(), PIP_LIST_PATH)


def get_wildcards(extensions):
    """

    Parameters
    ----------
    extensions :


    Returns
    -------

    """
    if type(extensions) is not list:
        extensions = [extensions]
    return "|".join(
        ["%s (*.%s)|*.%s" % (ext.upper(), ext, ext) for ext in extensions]
    )


FIG_WILDCARDS = get_wildcards(["svg", "html", "png"])


def set_phantom_js_in_path():
    """ """
    bundle_dir = getattr(sys, "_MEIPASS", None)
    phantom_js_path = APP_DIR if bundle_dir is None else bundle_dir

    if phantom_js_path not in environ["PATH"]:
        environ["PATH"] += pathsep + phantom_js_path


def backup_sqlite_db(options):
    """

    Parameters
    ----------
    options :


    Returns
    -------

    """
    if options.DB_TYPE == "sqlite":
        db_file_name = basename(options.DEFAULT_CNF["sqlite"]["host"])
        file_append = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        new_file_name = splitext(db_file_name)[0] + "_" + file_append + ".db"
        db_file_path = join(DATA_DIR, db_file_name)

        # check if stored file_path is an absolute path
        if not isfile(db_file_path):
            db_file_path = options.DEFAULT_CNF["sqlite"]["host"]
            if not isfile(db_file_path):
                db_file_path = None

        if db_file_path is not None:
            move_files_to_new_path(
                [db_file_path],
                BACKUP_DIR,
                copy_files=True,
                new_file_names=[new_file_name],
            )

            return [db_file_path, join(BACKUP_DIR, new_file_name)]


def main_is_frozen():
    """ """
    # https://pyinstaller.readthedocs.io/en/stable/runtime-information.html
    return getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS")


def get_xy_path_lengths(shapely_object):
    """Get the x and y path lengths of a a Shapely object

    Parameters
    ----------
    shapely_object :
        either 'GeometryCollection', 'MultiPolygon', or 'Polygon'

    Returns
    -------
    list
        path lengths in the x and y directions

    """
    path = np.array([0.0, 0.0])
    if shapely_object.type == "GeometryCollection":
        for geometry in shapely_object.geoms:
            if geometry.type in {"MultiPolygon", "Polygon"}:
                path = np.add(path, get_xy_path_lengths(geometry))
    elif shapely_object.type == "MultiPolygon":
        for shape in shapely_object:
            path = np.add(path, get_xy_path_lengths(shape))
    elif shapely_object.type == "Polygon":
        x, y = np.array(shapely_object.exterior.xy[0]), np.array(
            shapely_object.exterior.xy[1]
        )
        path = np.array(
            [np.sum(np.abs(np.diff(x))), np.sum(np.abs(np.diff(y)))]
        )

    return path.tolist()


def recalculate_plan_complexities_from_beams():
    """ """
    with DVH_SQL() as cnx:
        uids = cnx.get_unique_values("Plans", "study_instance_uid")

        for uid in uids:
            try:
                condition = "study_instance_uid = '%s'" % uid
                beam_complexities = cnx.query(
                    "Beams", "fx_count, complexity, fx_grp_number", condition
                )
                complexity = {}
                fx_counts = {}
                for row in beam_complexities:
                    fx_count, beam_complexity, fx_group_number = tuple(row)
                    if fx_group_number not in complexity:
                        complexity[fx_group_number] = 0.0
                        fx_counts[fx_group_number] = fx_count
                    complexity[fx_group_number] += beam_complexity

                total_fx = float(sum([fx for fx in fx_counts.values()]))
                plan_complexity = (
                    sum(
                        [
                            c * fx_counts[fx_grp]
                            for fx_grp, c in complexity.items()
                        ]
                    )
                    / total_fx
                )
            except Exception as e:
                msg = (
                    "tools.utilities.recalculate_plan_complexities_from_beams: failed on uid = %s"
                    % uid
                )
                push_to_log(e, msg=msg)
                plan_complexity = None

            if plan_complexity is not None:
                cnx.update("Plans", "complexity", plan_complexity, condition)


def get_new_uid(used_uids=None):
    """Get a new UID using pydicom

    Parameters
    ----------
    used_uids : list
        Do not return a UID that is in this list (Default value = None)

    Returns
    -------
    str
        A new UID not found in the current DVHA database or in the used_uids

    """
    uid_found = False
    used_uids = [] if used_uids is None else used_uids
    with DVH_SQL() as cnx:
        while not uid_found:
            uid = pydicom.uid.generate_uid()
            uid_found = not cnx.is_uid_imported(uid) and uid not in used_uids
    return uid


def get_new_uids_by_directory(start_path):
    """Generate new StudyInstanceUIDs, assuming all DICOM files in a directory are matched

    Parameters
    ----------
    start_path : str
        initial directory

    Returns
    -------
    dict, list
        New UIDs by directory, queue

    """
    file_paths = get_file_paths(
        start_path, search_subfolders=True, return_dict=True
    )
    study_uids = {}
    queue = []
    for directory, files in file_paths.items():
        uid = get_new_uid(used_uids=list(study_uids.values()))
        study_uids[directory] = uid
        for file in files:
            queue.append(
                {"abs_file_path": join(directory, file), "study_uid": uid}
            )

    return study_uids, queue


def edit_study_uid(abs_file_path, study_uid):
    """Change the StudyInstanceUID of a DICOM file

    Parameters
    ----------
    abs_file_path :
        absolute file path of the DICOM file
    study_uid :
        new StudyInstanceUID

    Returns
    -------

    """
    try:
        ds = pydicom.read_file(abs_file_path, force=True)
        ds.StudyInstanceUID = study_uid
        ds.save_as(abs_file_path)
    except Exception as e:
        push_to_log(e, abs_file_path)


def get_csv_row(data, columns, delimiter=","):
    """Convert a dictionary of data into a row for a csv file

    Parameters
    ----------
    data : dict
        a dictionary with values with str representations
    columns : list
        a list of keys dictating the order of the csv
    delimiter : str
        Optionally use the provided delimiter rather than a comma (Default value = ")
    " :


    Returns
    -------


    """
    str_data = [str(data[c]) for c in columns]
    clean_str_data = ['"%s"' % s if delimiter in s else s for s in str_data]
    clean_str_data = [s.replace("\n", "<>") for s in clean_str_data]
    return delimiter.join(clean_str_data)


def csv_to_list(csv_str, delimiter=","):
    """Split a CSV into a list

    Parameters
    ----------
    csv_str : str
        A comma-separated value string (with double quotes around values
        containing the delimiter)
    delimiter : str
        The str separator between values (Default value = ")
    " :


    Returns
    -------


    """
    if '"' not in csv_str:
        return csv_str.split(delimiter)

    # add an empty value with another ",", but ignore it
    # ensures next_csv_element always finds a ","
    next_value, csv_str = next_csv_element(csv_str + ",", delimiter)
    ans = [next_value.replace("<>", "\n")]
    while csv_str:
        next_value, csv_str = next_csv_element(csv_str, delimiter)
        ans.append(next_value.replace("<>", "\n"))

    return ans


def next_csv_element(csv_str, delimiter=","):
    """Helper function for csv_to_list

    Parameters
    ----------
    csv_str : str
        A comma-separated value string (with double quotes around values
        containing the delimiter)
    delimiter : str
        The str separator between values (Default value = ")
    " :


    Returns
    -------


    """
    if csv_str.startswith('"'):
        split = csv_str[1:].find('"') + 1
        return csv_str[1:split], csv_str[split + 2 :]

    next_delimiter = csv_str.find(delimiter)
    return csv_str[:next_delimiter], csv_str[next_delimiter + 1 :]
